from dataclasses import dataclass
from enum import StrEnum, auto
from typing import Any, cast

from eth_typing import ABIFunction
from eth_utils.abi import abi_to_signature, function_signature_to_4byte_selector
from lark import Lark, UnexpectedInput
from lark.visitors import Transformer_InPlaceRecursive

from erc7730.model.abi import ABI, Component, Function, InputOutput

_SIGNATURE_PARSER = parser = Lark(
    grammar=r"""
            function: identifier "(" params ")"
            
            params: (param ("," param)*)?
            ?param: named_param | named_tuple

            ?tuple: "(" params ")"
                       
            named_param: type identifier?
            named_tuple:  tuple identifier?

            array: "[]"
            identifier: /[a-zA-Z$_][a-zA-Z0-9$_]*/
            type: identifier array?

            %ignore " "
            """,
    start="function",
)


class FunctionTransformer(Transformer_InPlaceRecursive):
    """Visitor to transform the parsed function AST into function domain model objects."""

    def function(self, ast: Any) -> Function:
        (name, inputs) = ast
        return Function(
            name=name,
            inputs=[InputOutput(name=input.name, type=input.type, components=input.components) for input in inputs],
        )

    def params(self, ast: Any) -> list[Component]:
        return ast

    def named_param(self, ast: Any) -> Component:
        if len(ast) == 1:
            return Component(name="_", type=ast[0])
        (type_, name) = ast
        return Component(name=name, type=type_)

    def named_tuple(self, ast: Any) -> Component:
        if len(ast) == 1:
            return Component(name="_", type="tuple", components=ast[0])
        (components, name) = ast
        return Component(name=name, type="tuple", components=components)

    def array(self, ast: Any) -> str:
        return "[]"

    def identifier(self, ast: Any) -> str:
        (value,) = ast
        return value

    def type(self, ast: Any) -> str:
        if len(ast) == 1:
            return ast[0]

        (value, array) = ast
        return value + array


def compute_signature(abi: Function) -> str:
    """Compute the signature of a Function."""
    abi_function = cast(ABIFunction, abi.model_dump())
    return abi_to_signature(abi_function)


def reduce_signature(signature: str) -> str:
    """Remove parameter names and spaces from a function signature."""
    return compute_signature(parse_signature(signature))


def parse_signature(signature: str) -> Function:
    """Parse a function signature."""
    try:
        return FunctionTransformer().transform(_SIGNATURE_PARSER.parse(signature))
    except UnexpectedInput as e:
        raise ValueError(f"Invalid signature: {signature}") from e


def signature_to_selector(signature: str) -> str:
    """Compute the keccak of a signature."""
    return "0x" + function_signature_to_4byte_selector(signature).hex()


def function_to_selector(abi: Function) -> str:
    """Compute the selector of a Function."""
    return signature_to_selector(compute_signature(abi))


@dataclass(kw_only=True)
class Functions:
    functions: dict[str, Function]
    proxy: bool


def get_functions(abis: list[ABI]) -> Functions:
    """Get the functions from a list of ABIs."""
    functions = Functions(functions={}, proxy=False)
    for abi in abis:
        if abi.type == "function":
            functions.functions[function_to_selector(abi)] = abi
            if abi.name in ("proxyType", "getImplementation", "implementation", "proxy__getImplementation"):
                functions.proxy = True
    return functions


class ABIDataType(StrEnum):
    """Solidity data type."""

    UINT = auto()
    INT = auto()
    UFIXED = auto()
    FIXED = auto()
    ADDRESS = auto()
    BOOL = auto()
    BYTES = auto()
    STRING = auto()



from typing import final, override

from erc7730.common.abi import function_to_selector
from erc7730.common.output import OutputAdder
from erc7730.lint import ERC7730Linter
from erc7730.model.paths import DataPath, Field
from erc7730.model.paths.path_ops import data_path_ends_with, path_starts_with, to_absolute
from erc7730.model.paths.path_schemas import (
    compute_abi_schema_paths,
    compute_eip712_schema_paths,
    compute_format_schema_paths,
)
from erc7730.model.resolved.context import EIP712Schema, ResolvedContractContext, ResolvedEIP712Context
from erc7730.model.resolved.descriptor import ResolvedERC7730Descriptor

AUTHORIZED_MISSING_DISPLAY_FIELDS = {
    Field(identifier="nonce"),
    Field(identifier="sigDeadline"),
}


@final
class ValidateDisplayFieldsLinter(ERC7730Linter):
    """
    - for each field of schema/ABI, check that there is a display field
    - for each field, check that display configuration is relevant with field type
    """

    @override
    def lint(self, descriptor: ResolvedERC7730Descriptor, out: OutputAdder) -> None:
        self._validate_eip712_paths(descriptor, out)
        self._validate_abi_paths(descriptor, out)

    @classmethod
    def _validate_eip712_paths(cls, descriptor: ResolvedERC7730Descriptor, out: OutputAdder) -> None:
        if isinstance(descriptor.context, ResolvedEIP712Context) and descriptor.context.eip712.schemas is not None:
            primary_types: set[str] = set()
            for schema in descriptor.context.eip712.schemas:
                if isinstance(schema, EIP712Schema):
                    primary_types.add(schema.primaryType)
                    if schema.primaryType not in schema.types:
                        out.error(
                            title="Invalid EIP712 Schema",
                            message=f"Primary type `{schema.primaryType}` is not present in schema types. Please make "
                            f"sure the EIP-712 schema includes a definition for the primary type.",
                        )
                        continue
                    if schema.primaryType not in descriptor.display.formats:
                        out.error(
                            title="Missing Display Format",
                            message=f"Schema primary type `{schema.primaryType}` must have a display format defined.",
                        )
                        continue
                    eip712_paths = compute_eip712_schema_paths(schema)
                    primary_type_format = descriptor.display.formats[schema.primaryType]
                    format_paths = compute_format_schema_paths(primary_type_format).data_paths

                    if (excluded := primary_type_format.excluded) is not None:
                        excluded_paths = [to_absolute(path) for path in excluded]
                    else:
                        excluded_paths = []

                    for path in eip712_paths - format_paths:
                        if any(path_starts_with(path, excluded_path) for excluded_path in excluded_paths):
                            continue

                        if any(data_path_ends_with(path, allowed) for allowed in AUTHORIZED_MISSING_DISPLAY_FIELDS):
                            out.debug(
                                title="Optional Display field missing",
                                message=f"No display field is defined for path `{path}` in message "
                                f"{schema.primaryType}. If intentionally excluded, please add it to `excluded` "
                                f"list to avoid this warning.",
                            )
                        else:
                            out.warning(
                                title="Missing Display field",
                                message=f"No display field is defined for path `{path}` in {schema.primaryType}. "
                                f"If intentionally excluded, please add it to `excluded` list to avoid this "
                                f"warning.",
                            )
                    for path in format_paths - eip712_paths:
                        out.error(
                            title="Invalid Display field",
                            message=f"A display field is defined for `{path}`, but it does not exist in message "
                            f"{schema.primaryType}. Please check the field path is valid according to the EIP-712 "
                            f"schema.",
                        )

                else:
                    out.error(
                        title="Missing EIP712 Schema",
                        message=f"EIP712 Schema is missing (found {schema})",
                    )

            for fmt in descriptor.display.formats:
                if fmt not in primary_types:
                    out.error(
                        title="Invalid Display Format",
                        message=f"Type `{fmt}` is not in EIP712 schemas. Please check the type is valid according to "
                        f"the EIP-712 schema.",
                    )

    @classmethod
    def _validate_abi_paths(cls, descriptor: ResolvedERC7730Descriptor, out: OutputAdder) -> None:
        if isinstance(descriptor.context, ResolvedContractContext):
            abi_paths_by_selector: dict[str, set[DataPath]] = {}
            for abi in descriptor.context.contract.abi:
                if abi.type == "function":
                    abi_paths_by_selector[function_to_selector(abi)] = compute_abi_schema_paths(abi)

            for selector, fmt in descriptor.display.formats.items():
                if selector not in abi_paths_by_selector:
                    out.error(
                        title="Invalid selector",
                        message=f"Selector {selector} not found in ABI.",
                    )
                    continue
                format_paths = compute_format_schema_paths(fmt).data_paths
                abi_paths = abi_paths_by_selector[selector]

                if (excluded := fmt.excluded) is not None:
                    excluded_paths = [to_absolute(path) for path in excluded]
                else:
                    excluded_paths = []

                for path in abi_paths - format_paths:
                    if any(path_starts_with(path, excluded_path) for excluded_path in excluded_paths):
                        continue

                    if any(data_path_ends_with(path, allowed) for allowed in AUTHORIZED_MISSING_DISPLAY_FIELDS):
                        out.debug(
                            title="Optional Display Field Missing",
                            message=f"No display field is defined for path `{path}` in function {selector}. If "
                            f"intentionally excluded, please add it to `excluded` list to avoid this warning.",
                        )
                    else:
                        out.warning(
                            title="Missing Display field",
                            message=f"No display field is defined for path `{path}` in function {selector}. If "
                            f"intentionally excluded, please add it to `excluded` list to avoid this warning.",
                        )
                for path in format_paths - abi_paths:
                    out.error(
                        title="Invalid Display field",
                        message=f"A display field is defined for `{path}`, but it does not exist in function "
                        f"{selector} ABI. Please check the field path is valid according to the ABI.",
                    )



from typing import assert_never

from pydantic import TypeAdapter, ValidationError

from erc7730.common.abi import ABIDataType
from erc7730.common.output import OutputAdder
from erc7730.convert.resolved.constants import ConstantProvider
from erc7730.model.display import FieldFormat
from erc7730.model.input.display import InputFieldBase
from erc7730.model.paths import ContainerPath, DataPath, DescriptorPath
from erc7730.model.paths.path_ops import data_or_container_path_concat
from erc7730.model.resolved.display import ResolvedValue, ResolvedValueConstant, ResolvedValuePath
from erc7730.model.types import HexStr, ScalarType


def resolve_field_value(
    prefix: DataPath,
    input_field: InputFieldBase,
    input_field_format: FieldFormat | None,
    constants: ConstantProvider,
    out: OutputAdder,
) -> ResolvedValue | None:
    """
    Resolve value, as a data path or constant value, for a field or reference.

    :param prefix: current path prefix
    :param input_field: field description or definition
    :param input_field_format: input field format
    :param constants: descriptor paths constants resolver
    :param out: error handler
    :return: resolved value or None if error
    """
    match input_field_format:
        case None | FieldFormat.RAW:
            abi_type = ABIDataType.STRING
        case (
            FieldFormat.AMOUNT
            | FieldFormat.TOKEN_AMOUNT
            | FieldFormat.DURATION
            | FieldFormat.DATE
            | FieldFormat.UNIT
            | FieldFormat.NFT_NAME
            | FieldFormat.ENUM
        ):
            abi_type = ABIDataType.UINT
        case FieldFormat.ADDRESS_NAME:
            abi_type = ABIDataType.ADDRESS
        case FieldFormat.CALL_DATA:
            abi_type = ABIDataType.BYTES
        case _:
            assert_never(input_field_format)

    if (
        value := resolve_path_or_constant_value(
            prefix=prefix,
            input_path=input_field.path,
            input_value=input_field.value,
            abi_type=abi_type,
            constants=constants,
            out=out,
        )
    ) is None:
        return out.error(title="Invalid field", message="Field must have either a path or a value.")
    return value


def resolve_path_or_constant_value(
    prefix: DataPath,
    input_path: DescriptorPath | DataPath | ContainerPath | None,
    input_value: DescriptorPath | ScalarType | None,
    abi_type: ABIDataType,
    constants: ConstantProvider,
    out: OutputAdder,
) -> ResolvedValue | None:
    """
    Resolve value, as a data path or constant value.

    :param prefix: current path prefix
    :param input_path: input data path, if provided
    :param input_value: input constant value, if provided
    :param abi_type: expected encoded value data type
    :param constants: descriptor paths constants resolver
    :param out: error handler
    :return: resolved value or None if error or value resolves to None
    """
    if input_path is not None:
        if input_value is not None:
            return out.error(
                title="Invalid field",
                message="Field cannot have both a path and a value.",
            )

        if (path := constants.resolve_path(input_path, out)) is None:
            return None

        return ResolvedValuePath(path=data_or_container_path_concat(prefix, path))

    if input_value is not None:
        if (value := constants.resolve(input_value, out)) is None:
            return None

        if not isinstance(value, str | bool | int | float):
            return out.error(
                title="Invalid constant value",
                message="Constant value must be a scalar type (string, boolean or number).",
            )

        if (raw := encode_value(value, abi_type, out)) is None:
            return None

        return ResolvedValueConstant(type_family=abi_type, type_size=len(raw) // 2 - 1, value=value, raw=raw)

    return None


def encode_value(value: ScalarType, abi_type: ABIDataType, out: OutputAdder) -> HexStr | None:
    if isinstance(value, str) and value.startswith("0x"):
        try:
            return TypeAdapter(HexStr).validate_strings(value)
        except ValidationError:
            return out.error(
                title="Invalid hex string",
                message=f""""{value}" is not a valid hexadecimal string.""",
            )

    # this uses a custom specific encoding because this is what the Ledger app expects
    try:
        match abi_type:
            case ABIDataType.UFIXED | ABIDataType.FIXED:
                return out.error(title="Invalid constant", message="""Fixed precision numbers are not supported""")

            case ABIDataType.UINT:
                if not isinstance(value, int) or value < 0:
                    return out.error(title="Invalid constant", message=f"""Value "{value}" is not an unsigned int""")
                encoded = value.to_bytes(length=(max(value.bit_length(), 1) + 7) // 8, signed=False)

            case ABIDataType.INT:
                if not isinstance(value, int):
                    return out.error(title="Invalid constant", message=f"""Value "{value}" is not an integer""")
                encoded = value.to_bytes(length=(max(value.bit_length(), 1) + 7) // 8, signed=True)

            case ABIDataType.BOOL:
                if not isinstance(value, bool):
                    return out.error(title="Invalid constant", message=f"""Value "{value}" is not a boolean""")
                encoded = value.to_bytes()

            case ABIDataType.STRING:
                if not isinstance(value, str):
                    return out.error(title="Invalid constant", message=f"""Value "{value}" is not a string""")
                encoded = value.encode(encoding="ascii", errors="replace")

            case ABIDataType.ADDRESS:
                return out.error(
                    title="Invalid constant", message=f"""Value "{value}" is not a valid address hexadecimal string."""
                )

            case ABIDataType.BYTES:
                return out.error(
                    title="Invalid constant", message=f"""Value "{value}" is not a valid hexadecimal string."""
                )

            case _:
                assert_never(abi_type)
    except OverflowError:
        return out.error(
            title="Invalid constant",
            message=f"""Value "{value}" is too large for the specified type.""",
        )

    return HexStr("0x" + encoded.hex())



from dataclasses import dataclass
from typing import assert_never

from eip712.model.schema import EIP712SchemaField

from erc7730.model.abi import Component, Function, InputOutput
from erc7730.model.context import EIP712Schema
from erc7730.model.paths import (
    ROOT_DATA_PATH,
    Array,
    ArrayElement,
    ArraySlice,
    ContainerPath,
    DataPath,
    DataPathElement,
    Field,
)
from erc7730.model.paths.path_ops import data_path_append
from erc7730.model.resolved.display import (
    ResolvedAddressNameParameters,
    ResolvedCallDataParameters,
    ResolvedDateParameters,
    ResolvedEnumParameters,
    ResolvedField,
    ResolvedFieldDescription,
    ResolvedFormat,
    ResolvedNestedFields,
    ResolvedNftNameParameters,
    ResolvedTokenAmountParameters,
    ResolvedUnitParameters,
    ResolvedValue,
    ResolvedValueConstant,
    ResolvedValuePath,
)
from erc7730.model.resolved.path import ResolvedPath


@dataclass(kw_only=True, frozen=True)
class FormatPaths:
    data_paths: set[DataPath]  # References to values in the serialized data
    container_paths: set[ContainerPath]  # References to values in the container


def compute_eip712_schema_paths(schema: EIP712Schema) -> set[DataPath]:
    """
    Compute the sets of valid schema paths for an EIP-712 schema.

    :param schema: EIP-712 schema
    :return: valid schema paths
    """

    if (primary_type := schema.types.get(schema.primaryType)) is None:
        raise ValueError(f"Invalid schema: primaryType {schema.primaryType} not in types")

    paths: set[DataPath] = set()

    def append_paths(path: DataPath, current_type: list[EIP712SchemaField]) -> None:
        for field in current_type:
            if len(field.name) == 0:
                continue  # skip unnamed parameters

            sub_path = data_path_append(path, Field(identifier=field.name))

            field_base_type = field.type.rstrip("[]")

            if field_base_type in {"bytes"}:
                paths.add(data_path_append(sub_path, Array()))

            if field_base_type != field.type:
                sub_path = data_path_append(sub_path, Array())
                paths.add(sub_path)

            if (target_type := schema.types.get(field_base_type)) is not None:
                append_paths(sub_path, target_type)
            else:
                paths.add(sub_path)

    append_paths(ROOT_DATA_PATH, primary_type)

    return paths


def compute_abi_schema_paths(abi: Function) -> set[DataPath]:
    """
    Compute the sets of valid schema paths for an ABI function.

    :param abi: Solidity ABI function
    :return: valid schema paths
    """
    paths: set[DataPath] = set()

    def append_paths(path: DataPath, params: list[InputOutput] | list[Component] | None) -> None:
        if not params:
            return None
        for param in params:
            if len(param.name) == 0:
                continue  # skip unnamed parameters

            sub_path = data_path_append(path, Field(identifier=param.name))

            param_base_type = param.type.rstrip("[]")

            if param_base_type in {"bytes"}:
                paths.add(data_path_append(sub_path, Array()))

            if param_base_type != param.type:
                sub_path = data_path_append(sub_path, Array())
                paths.add(sub_path)

            if param.components:
                append_paths(sub_path, param.components)  # type: ignore
            else:
                paths.add(sub_path)

    append_paths(ROOT_DATA_PATH, abi.inputs)

    return paths


def compute_format_schema_paths(format: ResolvedFormat) -> FormatPaths:
    """
    Compute the sets of schema paths referred in an ERC7730 Format section.

    :param format: resolved $.display.format section
    :return: schema paths used by field formats
    """
    data_paths: set[DataPath] = set()  # references to values in the serialized data
    container_paths: set[ContainerPath] = set()  # references to values in the container

    if format.fields is not None:

        def add_path(path: ResolvedPath | None) -> None:
            match path:
                case None:
                    pass
                case ContainerPath():
                    container_paths.add(path)
                case DataPath():
                    data_paths.add(data_path_to_schema_path(path))
                case _:
                    assert_never(path)

        def add_value(value: ResolvedValue | None) -> None:
            match value:
                case None:
                    pass
                case ResolvedValueConstant():
                    pass
                case ResolvedValuePath(path=path):
                    add_path(path)
                case _:
                    assert_never(value)

        def append_paths(field: ResolvedField) -> None:
            add_value(field.value)
            match field:
                case ResolvedFieldDescription():
                    match field.params:
                        case None:
                            pass
                        case ResolvedAddressNameParameters():
                            pass
                        case ResolvedCallDataParameters(callee=callee):
                            add_value(callee)
                        case ResolvedTokenAmountParameters(token=token):
                            add_value(token)
                        case ResolvedNftNameParameters(collection=collection):
                            add_value(collection)
                        case ResolvedDateParameters():
                            pass
                        case ResolvedUnitParameters():
                            pass
                        case ResolvedEnumParameters():
                            pass
                        case _:
                            assert_never(field.params)
                case ResolvedNestedFields():
                    for nested_field in field.fields:
                        append_paths(nested_field)
                case _:
                    assert_never(field)

        for field in format.fields:
            append_paths(field)

    return FormatPaths(data_paths=data_paths, container_paths=container_paths)


def data_path_to_schema_path(path: DataPath) -> DataPath:
    """
    Convert a data path to a schema path.

    Example: #.foo.[].[-2].bar.[1:5] -> #.foo.[].[].bar

    :param path: data path
    :return: schema path
    """

    def to_schema(element: DataPathElement) -> DataPathElement | None:
        match element:
            case Field() as f:
                return f
            case Array() | ArrayElement():
                return Array()
            # TODO: Spec also allows slicing on array type, but for now it is only used on primitive types
            case ArraySlice():
                return None
            case _:
                assert_never(element)

    return path.model_copy(update={"elements": [to_schema(e) for e in path.elements if to_schema(e) is not None]})



from pathlib import Path

import pytest
from eip712.model.input.descriptor import InputEIP712DAppDescriptor

from erc7730.common.json import dict_from_json_file
from erc7730.common.pydantic import model_to_json_dict
from erc7730.convert.convert import convert_and_print_errors
from erc7730.convert.ledger.eip712.convert_erc7730_to_eip712 import ERC7730toEIP712Converter
from erc7730.convert.resolved.convert_erc7730_input_to_resolved import ERC7730InputToResolved
from erc7730.model.input.descriptor import InputERC7730Descriptor
from erc7730.model.resolved.descriptor import ResolvedERC7730Descriptor
from tests.assertions import assert_dict_equals
from tests.cases import path_id
from tests.files import ERC7730_EIP712_DESCRIPTORS
from tests.schemas import assert_valid_legacy_eip_712
from tests.skip import single_or_first, single_or_skip

DATA = Path(__file__).resolve().parent / "data"


@pytest.mark.parametrize("input_file", ERC7730_EIP712_DESCRIPTORS, ids=path_id)
def test_erc7730_registry_files(input_file: Path) -> None:
    """
    Test converting ERC-7730 => Ledger legacy EIP-712.

    Note the test only applies to descriptors with a single contract and message, and only checks output files are
    compliant with the Ledger legacy EIP-712 json schema.
    """
    input_erc7730_descriptor = InputERC7730Descriptor.load(input_file)
    resolved_erc7730_descriptor = convert_and_print_errors(input_erc7730_descriptor, ERC7730InputToResolved())
    resolved_erc7730_descriptor = single_or_skip(resolved_erc7730_descriptor)
    output_descriptor = convert_and_print_errors(resolved_erc7730_descriptor, ERC7730toEIP712Converter())
    output_descriptor = single_or_skip(output_descriptor)
    assert_valid_legacy_eip_712(output_descriptor)


@pytest.mark.parametrize("input_file", ERC7730_EIP712_DESCRIPTORS, ids=path_id)
def test_erc7730_registry_files_by_reference(input_file: Path) -> None:
    """
    Test converting ERC-7730 => Ledger legacy EIP-712.

    Note the test only applies to descriptors with a single contract and message, and only checks output files are
    compliant with the Ledger legacy EIP-712 json schema.
    """
    reference_path = DATA / input_file.name
    if not reference_path.is_file():
        pytest.skip(f"No reference file at {reference_path}")
    input_erc7730_descriptor = InputERC7730Descriptor.load(input_file)
    resolved_erc7730_descriptors = convert_and_print_errors(input_erc7730_descriptor, ERC7730InputToResolved())
    resolved_erc7730_descriptor: ResolvedERC7730Descriptor = single_or_first(resolved_erc7730_descriptors)
    output_descriptors = convert_and_print_errors(resolved_erc7730_descriptor, ERC7730toEIP712Converter())
    output_descriptor: InputEIP712DAppDescriptor = single_or_first(output_descriptors)
    assert_dict_equals(dict_from_json_file(reference_path), model_to_json_dict(output_descriptor))



import json
import logging
import os
from pathlib import Path
from typing import Annotated, assert_never

from eip712.convert.input_to_resolved import EIP712InputToResolvedConverter
from eip712.model.input.descriptor import InputEIP712DAppDescriptor
from pydantic_string_url import HttpUrl
from rich import print
from typer import Argument, Exit, Option, Typer

from erc7730.common.output import ConsoleOutputAdder
from erc7730.convert.convert import convert_to_file_and_print_errors
from erc7730.convert.ledger.eip712.convert_eip712_to_erc7730 import EIP712toERC7730Converter
from erc7730.convert.ledger.eip712.convert_erc7730_to_eip712 import ERC7730toEIP712Converter
from erc7730.convert.resolved.convert_erc7730_input_to_resolved import ERC7730InputToResolved
from erc7730.format.format import format_all_and_print_errors
from erc7730.generate.generate import generate_descriptor
from erc7730.lint.lint import lint_all_and_print_errors
from erc7730.list.list import list_all
from erc7730.model import ERC7730ModelType
from erc7730.model.base import Model
from erc7730.model.input.descriptor import InputERC7730Descriptor
from erc7730.model.resolved.descriptor import ResolvedERC7730Descriptor
from erc7730.model.types import Address

if os.environ.get("DEBUG") is not None:
    logging.basicConfig(
        format="%(levelname)s [%(asctime)s] %(name)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S", level=logging.DEBUG
    )


app = Typer(
    name="erc7730",
    no_args_is_help=True,
    help="""
    ERC-7730 tool.
    """,
)
convert_app = Typer(
    name="convert",
    no_args_is_help=True,
    short_help="Commands to convert descriptor files.",
    help="""
    Commands to convert descriptor files.
    """,
)
app.add_typer(convert_app)


@app.command(
    name="schema",
    short_help="Print ERC-7730 descriptor JSON schema.",
    help="""
    Print ERC-7730 descriptor JSON schema.
    """,
)
def command_schema(
    model_type: Annotated[ERC7730ModelType, Argument(help="The descriptor form ")] = ERC7730ModelType.INPUT,
) -> None:
    descriptor_type: type[Model]
    match model_type:
        case ERC7730ModelType.INPUT:
            descriptor_type = InputERC7730Descriptor
        case ERC7730ModelType.RESOLVED:
            descriptor_type = ResolvedERC7730Descriptor
        case _:
            assert_never(model_type)

    print(json.dumps(descriptor_type.model_json_schema(by_alias=True), indent=4))


@app.command(
    name="lint",
    short_help="Validate descriptor files.",
    help="""
    Validate descriptor files.
    """,
)
def command_lint(
    paths: Annotated[list[Path], Argument(help="The files or directory paths to lint")],
    gha: Annotated[bool, Option(help="Enable Github annotations output")] = False,
) -> None:
    if not lint_all_and_print_errors(paths, gha):
        raise Exit(1)


@app.command(
    name="list",
    short_help="List descriptor files.",
    help="""
    Recursively list all descriptor files, starting from current directory by default.
    """,
)
def command_list(
    paths: Annotated[list[Path] | None, Argument(help="The files or directory paths to search")] = None,
) -> None:
    if not list_all(paths or [Path.cwd()]):
        raise Exit(1)


@app.command(
    name="format",
    short_help="Format descriptor files.",
    help="""
    Recursively find and format all descriptor files, starting from current directory by default.
    """,
)
def command_format(
    paths: Annotated[list[Path] | None, Argument(help="The files or directory paths to search")] = None,
) -> None:
    if not format_all_and_print_errors(paths or [Path.cwd()]):
        raise Exit(1)


@app.command(
    name="resolve",
    short_help="Convert descriptor to resolved form.",
    help="""
    Convert descriptor to resolved form:
        - URLs fetched
        - Contract addresses normalized to lowercase
        - References inlined
        - Constants inlined
        - Field definitions inlined
        - Nested fields flattened where possible
        - Selectors converted to 4 bytes form
    
    See `erc7730 schema resolved` for the resolved descriptor schema.
    """,
)
def command_resolve(
    input_path: Annotated[Path, Argument(help="The input ERC-7730 file path")],
) -> None:
    input_descriptor = InputERC7730Descriptor.load(input_path)
    if (resolved_descriptor := ERC7730InputToResolved().convert(input_descriptor, ConsoleOutputAdder())) is None:
        raise Exit(1)
    print(resolved_descriptor.to_json_string())


@app.command(
    name="generate",
    short_help="Bootstrap a descriptor for given ABI/schema.",
    help="""
    Fetches ABI or schema files and generates a minimal descriptor.
    """,
)
def command_generate(
    chain_id: Annotated[int, Option(help="The EIP-155 chain id")],
    address: Annotated[Address, Option(help="The contract address")],
    abi: Annotated[Path | None, Option(help="Path to a JSON ABI file (to generate a calldata descriptor)")] = None,
    schema: Annotated[Path | None, Option(help="Path to an EIP-712 schema (to generate an EIP-712 descriptor)")] = None,
    owner: Annotated[str | None, Option(help="The display name of the owner or target of the contract")] = None,
    legal_name: Annotated[str | None, Option(help="The full legal name of the owner")] = None,
    url: Annotated[str | None, Option(help="URL with more info on the entity interacted with")] = None,
) -> None:
    if schema is not None and abi is not None:
        print("Cannot specify both ABI and schema.")
        raise Exit(1)
    schema_buffer = None
    abi_buffer = None

    if schema is not None:
        with open(schema, "rb") as f:
            schema_buffer = f.read()
    elif abi is not None:
        with open(abi, "rb") as f:
            abi_buffer = f.read()

    descriptor = generate_descriptor(
        chain_id=chain_id,
        contract_address=address,
        abi=abi_buffer,
        eip712_schema=schema_buffer,
        owner=owner,
        legal_name=legal_name,
        url=HttpUrl(url) if url is not None else None,
    )
    print(descriptor.to_json_string())


@convert_app.command(
    name="eip712-to-erc7730",
    short_help="Convert a legacy EIP-712 descriptor file to an ERC-7730 file.",
    help="""
    Convert a legacy EIP-712 descriptor file to an ERC-7730 file.
    """,
)
def command_convert_eip712_to_erc7730(
    input_eip712_path: Annotated[Path, Argument(help="The input EIP-712 file path")],
    output_erc7730_path: Annotated[Path, Argument(help="The output ERC-7730 file path")],
) -> None:
    input_descriptor = InputEIP712DAppDescriptor.load(input_eip712_path)
    resolved_descriptor = EIP712InputToResolvedConverter().convert(input_descriptor)
    if not convert_to_file_and_print_errors(
        input_descriptor=resolved_descriptor,
        output_file=output_erc7730_path,
        converter=EIP712toERC7730Converter(),
    ):
        raise Exit(1)


@convert_app.command(
    name="erc7730-to-eip712",
    short_help="Convert an ERC-7730 file to a legacy EIP-712 descriptor file.",
    help="""
    Convert an ERC-7730 file to a legacy EIP-712 descriptor file (if applicable).
    """,
)
def command_convert_erc7730_to_eip712(
    input_erc7730_path: Annotated[Path, Argument(help="The input ERC-7730 file path")],
    output_eip712_path: Annotated[Path, Argument(help="The output EIP-712 file path")],
) -> None:
    input_descriptor = InputERC7730Descriptor.load(input_erc7730_path)
    resolved_descriptor = ERC7730InputToResolved().convert(input_descriptor, ConsoleOutputAdder())
    if resolved_descriptor is None or not convert_to_file_and_print_errors(
        input_descriptor=resolved_descriptor,
        output_file=output_eip712_path,
        converter=ERC7730toEIP712Converter(),
    ):
        raise Exit(1)


if __name__ == "__main__":
    app()



#!/usr/bin/env python3
import glob
import logging
from pathlib import Path
from typing import Callable, Iterable, Tuple

from jsonschema.exceptions import ValidationError

from validation.validators import (
    abi_filename_validator,
    eip712_schema_validator,
    endlines_validator,
    format_validator,
    missing_abi_validator,
    run_validations,
    check_duplicate_contract,
    schema_validator,
    unique_field_validator,
    contract_matching_validator,
    missing_schema_validator,
    eip55_address_validator,
    check_duplicate_plugin,
)

logging.basicConfig(level="INFO")


def schema_validator_generator() -> Iterable[Tuple[str, Tuple[str, Callable]]]:
    json_schema_paths = "./*/*.schema.json"
    for json_schema in glob.iglob(json_schema_paths, recursive=True):
        schema_path = Path(json_schema)
        coin_name = schema_path.parent.name
        target_name = schema_path.name.removesuffix(".schema.json")
        yield f"{coin_name}_{target_name}", (
            str(schema_path.parent / "*" / f"{target_name}.json"),
            schema_validator(str(schema_path)),
        )


SCHEMA_VALIDATORS = {
    validator_name: validator
    for validator_name, validator in schema_validator_generator()
}

VALIDATORS = {
    "json": ("./*/**/*.json", format_validator),
    "endlines": ("./*/**/*.json", endlines_validator),
    "eip712_schema": ("./*/**/eip712.json", eip712_schema_validator),
    "abis_format": ("./*/**/abis/*.json", abi_filename_validator),
    "unique_id": ("./*/**/parsers.json", unique_field_validator(["id"])),
    "parsers_abi_not_missing": ("./*/**/parsers.json", missing_abi_validator),
    "b2c_abi_not_missing": ("./*/**/b2c.json", missing_abi_validator),
    "parsers_schema_not_missing": ("./*/**/parsers.json", missing_schema_validator()),
    "b2c_schema_not_missing": ("./*/**/b2c.json", missing_schema_validator()),
    "eip55_parser": ("./*/**/parsers.json", eip55_address_validator),
    "contract_matching": ("./*/**/parsers.json", contract_matching_validator),
    **SCHEMA_VALIDATORS,
}

if __name__ == "__main__":
    failed = False
    logging.basicConfig(level=logging.INFO, format='%(message)s', force=True)
    logger = logging.getLogger(__name__)

    try:
        logger.info("Running duplicate contract check")
        check_duplicate_contract("./*/**/b2c.json")
        logger.info("Running duplicate plugin check")
        check_duplicate_plugin("./*/**/b2c.json")
    except ValidationError:
        failed = True

    for validator_name, (path, validator) in VALIDATORS.items():
        logger.info("Running validation for %s", validator_name)
        try:
            run_validations(path, validator)
        except ValidationError:
            failed = True

    if failed:
        exit(1)



import json
from pathlib import Path
from shutil import copytree

import pytest
from typer.testing import CliRunner

from erc7730.main import app
from erc7730.model import ERC7730ModelType
from tests.cases import path_id
from tests.files import (
    ERC7730_DESCRIPTORS,
    ERC7730_EIP712_DESCRIPTORS,
    ERC7730_REGISTRY_ROOT,
    LEGACY_EIP712_DESCRIPTORS,
)

runner = CliRunner()


def test_help() -> None:
    result = runner.invoke(app, ["--help"])
    out = "".join(result.stdout.splitlines())
    assert "ERC-7730" in out
    assert "convert" in out
    assert "lint" in out


@pytest.mark.parametrize("model_type", list(ERC7730ModelType))
def test_schema(model_type: ERC7730ModelType) -> None:
    result = runner.invoke(app, ["schema", model_type])
    out = "".join(result.stdout.splitlines())
    assert result.exit_code == 0
    assert json.loads(out) is not None


def test_list() -> None:
    result = runner.invoke(app, ["list", str(ERC7730_REGISTRY_ROOT)])
    out = "".join(result.stdout.splitlines())
    assert "calldata-" in out
    assert "eip712-" in out
    assert ".json" in out


def test_format(tmp_path: Path) -> None:
    copytree(ERC7730_REGISTRY_ROOT, tmp_path / "registry")
    result = runner.invoke(app, ["format", str(tmp_path)])
    out = "".join(result.stdout.splitlines())
    assert "calldata-" in out
    assert "eip712-" in out
    assert ".json" in out
    assert "no errors occurred ✅" in out


@pytest.mark.parametrize("input_file", ERC7730_DESCRIPTORS, ids=path_id)
def test_lint_registry_files(input_file: Path) -> None:
    result = runner.invoke(app, ["lint", str(input_file)])
    out = "".join(result.stdout.splitlines())
    assert str(input_file.name) in out
    assert any(
        (
            "no errors found ✅" in out,
            "some warnings found ⚠️" in out,
            "some errors found ❌" in out,
        )
    )


@pytest.mark.parametrize("input_file", ERC7730_DESCRIPTORS, ids=path_id)
def test_resolve_registry_files(input_file: Path) -> None:
    result = runner.invoke(app, ["resolve", str(input_file)])
    out = "".join(result.stdout.splitlines())
    assert json.loads(out) is not None


@pytest.mark.parametrize(
    "label,chain_id,contract_address",
    [
        ("Uniswap v3 Router 2", 8453, "0x3fC91A3afd70395Cd496C647d5a6CC9D4B2b7FAD"),
    ],
)
def test_generate_from_contract_address(label: str, chain_id: int, contract_address: str) -> None:
    result = runner.invoke(app, ["generate", f"--chain-id={chain_id}", f"--address={contract_address}"])
    out = "".join(result.stdout.splitlines())
    assert json.loads(out) is not None


@pytest.mark.parametrize("input_file", LEGACY_EIP712_DESCRIPTORS, ids=path_id)
def test_convert_legacy_registry_eip712_files(input_file: Path, tmp_path: Path) -> None:
    result = runner.invoke(app, ["convert", "eip712-to-erc7730", str(input_file), str(tmp_path / input_file.name)])
    out = "".join(result.stdout.splitlines())
    assert "generated" in out
    assert "✅" in out


@pytest.mark.parametrize("input_file", ERC7730_EIP712_DESCRIPTORS, ids=path_id)
def test_convert_registry_files_to_legacy_eip712_files(input_file: Path, tmp_path: Path) -> None:
    result = runner.invoke(app, ["convert", "erc7730-to-eip712", str(input_file), str(tmp_path / input_file.name)])
    out = "".join(result.stdout.splitlines())
    assert "generated" in out
    assert "✅" in out



from abc import ABC
from typing import assert_never

import eth_abi
from eip712.model.schema import EIP712SchemaField, EIP712Type
from eth_abi.grammar import BasicType, TupleType
from pydantic import Field

from erc7730.common.abi import ABIDataType
from erc7730.model.abi import Component, Function, InputOutput
from erc7730.model.base import Model
from erc7730.model.context import EIP712Schema


class SchemaNode(Model, ABC):
    """Represents a node in the tree defined by a function ABI / EIP-712 schema."""


class SchemaStruct(SchemaNode):
    """Schema node representing a function or a tuple."""

    components: dict[str, "SchemaTree"] = Field(title="Struct components")


class SchemaArray(SchemaNode):
    """Schema node representing an array."""

    component: "SchemaTree" = Field(title="Array element type")


class SchemaLeaf(SchemaNode):
    """Schema node representing a scalar type."""

    data_type: ABIDataType = Field(title="Data type")


SchemaTree = SchemaStruct | SchemaArray | SchemaLeaf


def eip712_schema_to_tree(schema: EIP712Schema) -> SchemaTree:
    """
    Convert an EIP-712 schema to a schema tree.

    A schema tree is a tree representation of the EIP-712 schema, enriched with some metadata to ease crafting
    paths to access values in the message.

    :param schema: EIP-712 schema
    :return: Schema tree
    """
    if (primary_type_fields := schema.types.get(schema.primaryType)) is None:
        raise ValueError("Primary type not found in schema types")

    return _eip712_struct_type_to_tree(fields=primary_type_fields, types=schema.types)


def _eip712_struct_type_to_tree(
    fields: list[EIP712SchemaField], types: dict[EIP712Type, list[EIP712SchemaField]]
) -> SchemaTree:
    """
    Convert a complex EIP-712 type to a schema tree node (can be the primary type directly).

    :param fields: type fields
    :param types: all schema types
    :return: Schema tree
    """
    return SchemaStruct(components={field.name: _eip712_field_to_tree(field, types) for field in fields})


def _eip712_field_to_tree(field: EIP712SchemaField, types: dict[EIP712Type, list[EIP712SchemaField]]) -> SchemaTree:
    """
    Convert an EIP-712 type to a schema tree node.

    :param field: ABI element (can be a single component, or a function input)
    :param types: all schema types
    :return: Schema tree
    """
    match eth_abi.grammar.parse(field.type):
        case TupleType():
            return _eip712_struct_type_to_tree(types[field.type], types)

        case BasicType() as tp:
            if tp.is_array:
                match tp.base:
                    case "tuple" | "struct":
                        component = _eip712_struct_type_to_tree(types[field.type], types)
                    case _:
                        component = _eip712_field_to_tree(
                            field=field.model_copy(update={"type": tp.item_type.to_type_str()}), types=types
                        )

                return SchemaArray(component=component)

            match tp.base:
                case "tuple" | "struct":
                    return _eip712_struct_type_to_tree(types[field.type], types)
                case base if base in set(ABIDataType):
                    type_family = ABIDataType(base)
                case base_type:
                    if (base_type_fields := types.get(base_type)) is None:
                        raise Exception(f"Unexpected EIP-712 type: {base_type}")
                    return _eip712_struct_type_to_tree(base_type_fields, types)

            return SchemaLeaf(data_type=type_family)

        case unknown:
            raise Exception(f"Unexpected EIP-712 type: {type(unknown)}")


def abi_function_to_tree(function: Function) -> SchemaTree:
    """
    Convert a function ABI to a schema tree.

    A schema tree is a tree representation of the ABI of a function inputs, enriched with some metadata to ease
    crafting paths to access values in the serialized calldata.

    :param function: function ABI
    :return: Schema tree
    """
    return _abi_struct_component_to_tree(function)


def _abi_struct_component_to_tree(inp: Function | InputOutput | Component) -> SchemaTree:
    """
    Convert a struct-like ABI component to a schema tree node (can be the top level function directly).

    :param inp: ABI element
    :return: Schema tree
    """

    # get inputs/components list based on argument type
    input_components: list[InputOutput | Component] = []
    match inp:
        case Function(inputs=inputs):
            if inputs is not None:
                input_components.extend(inputs)
        case InputOutput(components=inp_components):
            if inp_components is not None:
                input_components.extend(inp_components)
        case Component(components=inp_components):
            if inp_components is not None:
                input_components.extend(inp_components)
        case list() as eip712_fields:
            input_components.extend(eip712_fields)
        case _:
            assert_never(inp)

    return SchemaStruct(
        components={component.name: _abi_component_to_tree(component) for component in input_components}
    )


def _abi_component_to_tree(inp: InputOutput | Component) -> SchemaTree:
    """
    Convert an ABI component to a schema tree node.

    :param inp: ABI element (can be a single component, or a function input)
    :return: Schema tree
    """
    match eth_abi.grammar.parse(inp.type):
        case TupleType():
            return _abi_struct_component_to_tree(inp)

        case BasicType() as tp:
            if tp.is_array:
                match tp.base:
                    case "tuple" | "struct":
                        component = _abi_struct_component_to_tree(inp)
                    case _:
                        component = _abi_component_to_tree(inp.model_copy(update={"type": tp.item_type.to_type_str()}))

                return SchemaArray(component=component)

            match tp.base:
                case "tuple" | "struct":
                    return _abi_struct_component_to_tree(inp)
                case base if base in set(ABIDataType):
                    type_family = ABIDataType(base)
                case unknown:
                    raise Exception(f"Unexpected ABI type: {unknown}")

            return SchemaLeaf(data_type=type_family)

        case unknown:
            raise Exception(f"Unexpected ABI type: {type(unknown)}")



"""
Object model for Solidity ABIs.

See https://docs.soliditylang.org/en/latest/abi-spec.html
"""

from enum import StrEnum
from typing import Annotated, Literal, Self

from pydantic import Field

from erc7730.model.base import Model

# ruff: noqa: N815 - camel case field names are tolerated to match schema


class StateMutability(StrEnum):
    pure = "pure"
    view = "view"
    nonpayable = "nonpayable"
    payable = "payable"


class Component(Model):
    name: str
    type: str
    internalType: str | None = None
    components: list[Self] | None = None


class InputOutput(Model):
    name: str
    type: str
    internalType: str | None = None
    components: list[Component] | None = None
    indexed: bool | None = None
    unit: str | None = None


class Function(Model):
    type: Literal["function"] = "function"
    name: str | None = None
    inputs: list[InputOutput] | None = None
    outputs: list[InputOutput] | None = None
    stateMutability: StateMutability | None = None
    constant: bool | None = None
    payable: bool | None = None
    gas: int | None = None
    signature: str | None = None


class Constructor(Model):
    type: Literal["constructor"] = "constructor"
    name: str | None = None
    inputs: list[InputOutput] | None = None
    outputs: list[InputOutput] | None = None
    stateMutability: StateMutability | None = None
    constant: bool | None = None
    payable: bool | None = None
    gas: int | None = None
    signature: str | None = None


class Receive(Model):
    type: Literal["receive"] = "receive"
    name: str | None = None
    inputs: list[InputOutput] | None = None
    outputs: list[InputOutput] | None = None
    stateMutability: StateMutability | None = None
    constant: bool | None = None
    payable: bool | None = None
    gas: int | None = None
    signature: str | None = None


class Fallback(Model):
    type: Literal["fallback"] = "fallback"
    name: str | None = None
    inputs: list[InputOutput] | None = None
    outputs: list[InputOutput] | None = None
    stateMutability: StateMutability | None = None
    constant: bool | None = None
    payable: bool | None = None
    gas: int | None = None
    signature: str | None = None


class Event(Model):
    type: Literal["event"] = "event"
    name: str
    inputs: list[InputOutput]
    anonymous: bool = False
    signature: str | None = None


class Error(Model):
    type: Literal["error"] = "error"
    name: str
    inputs: list[InputOutput]
    signature: str | None = None


ABI = Annotated[Constructor | Event | Function | Fallback | Error | Receive, Field(discriminator="type")]



import json
from pathlib import Path
from typing import Any

import pytest
from eip712.convert.input_to_resolved import EIP712InputToResolvedConverter
from eip712.model.input.descriptor import InputEIP712DAppDescriptor

from erc7730.common.pydantic import model_to_json_str
from erc7730.convert.convert import convert_and_print_errors
from erc7730.convert.ledger.eip712.convert_eip712_to_erc7730 import EIP712toERC7730Converter
from erc7730.convert.ledger.eip712.convert_erc7730_to_eip712 import ERC7730toEIP712Converter
from erc7730.convert.resolved.convert_erc7730_input_to_resolved import ERC7730InputToResolved
from erc7730.model.display import FieldFormat
from erc7730.model.input.descriptor import InputERC7730Descriptor
from tests.assertions import assert_dict_equals
from tests.cases import path_id
from tests.dict_utils import del_by_path, is_in_path, map_by_path
from tests.files import ERC7730_EIP712_DESCRIPTORS, LEGACY_EIP712_DESCRIPTORS
from tests.skip import single_or_skip


@pytest.mark.parametrize("input_file", ERC7730_EIP712_DESCRIPTORS, ids=path_id)
def test_roundtrip_from_erc7730(input_file: Path) -> None:
    """
    Test roundtrip conversion ERC-7730 => Ledger legacy EIP-712 => ERC-7730.

    Note the test only applies to descriptors with a single contract and message.
    """
    input_erc7730_descriptor = InputERC7730Descriptor.load(input_file)
    resolved_erc7730_descriptor = convert_and_print_errors(input_erc7730_descriptor, ERC7730InputToResolved())
    resolved_erc7730_descriptor = single_or_skip(resolved_erc7730_descriptor)
    legacy_eip712_descriptor = convert_and_print_errors(resolved_erc7730_descriptor, ERC7730toEIP712Converter())
    legacy_eip712_descriptor = single_or_skip(legacy_eip712_descriptor)
    resolved_legacy_eip172_descriptor = EIP712InputToResolvedConverter().convert(legacy_eip712_descriptor)
    output_erc7730_descriptor = convert_and_print_errors(resolved_legacy_eip172_descriptor, EIP712toERC7730Converter())
    output_erc7730_descriptor = single_or_skip(output_erc7730_descriptor)
    _assert_erc7730_json_equals_with_tolerance(input_erc7730_descriptor, output_erc7730_descriptor)


@pytest.mark.parametrize("input_file", LEGACY_EIP712_DESCRIPTORS, ids=path_id)
def test_roundtrip_from_legacy_eip712(input_file: Path) -> None:
    """
    Test roundtrip conversion Ledger legacy EIP-712 => ERC-7730 => Ledger legacy EIP-712.

    Note the test only applies to descriptors with a single contract and message.
    """

    if str(input_file).endswith("/uniswap/eip712.json"):
        pytest.skip("Several Uniswap messages have same primary types so they collide")

    input_legacy_eip712_descriptor = InputEIP712DAppDescriptor.load(input_file)
    resolved_legacy_eip172_descriptor = EIP712InputToResolvedConverter().convert(input_legacy_eip712_descriptor)
    input_erc7730_descriptor = convert_and_print_errors(resolved_legacy_eip172_descriptor, EIP712toERC7730Converter())
    input_erc7730_descriptor = single_or_skip(input_erc7730_descriptor)
    resolved_erc7730_descriptor = convert_and_print_errors(input_erc7730_descriptor, ERC7730InputToResolved())
    resolved_erc7730_descriptor = single_or_skip(resolved_erc7730_descriptor)
    output_legacy_eip712_descriptor = convert_and_print_errors(resolved_erc7730_descriptor, ERC7730toEIP712Converter())
    output_legacy_eip712_descriptor = single_or_skip(output_legacy_eip712_descriptor)
    _assert_eip712_json_equals_with_tolerance(input_legacy_eip712_descriptor, output_legacy_eip712_descriptor)


def _assert_erc7730_json_equals_with_tolerance(input: InputERC7730Descriptor, output: InputERC7730Descriptor) -> None:
    input_dict, output_dict = json.loads(model_to_json_str(input)), json.loads(model_to_json_str(output))

    # $schema is not present in EIP-712
    del_by_path(input_dict, "$schema")

    # Addresses may be in EIP-55 format, convert to lower case in input
    def _lowercase_addresses(deployments: list[dict[str, Any]]) -> Any:
        for deployment in deployments:
            if "address" in deployment:
                deployment["address"] = deployment["address"].lower()
        return deployments

    map_by_path(
        input_dict,
        _lowercase_addresses,
        "context",
        "eip712",
        "deployments",
    )

    # chainId and verifyingContract are always generated by the converter
    # even when not explicitly present in the input
    for key in "chainId", "verifyingContract":
        path = "context", "eip712", "domain", key
        if not is_in_path(input_dict, *path):
            del_by_path(output_dict, *path)

    # Some metadata cannot be stored in EIP712, and thus are lost with the roundtrip
    del_by_path(input_dict, "metadata", "info")
    del_by_path(input_dict, "metadata", "token")
    del_by_path(input_dict, "metadata", "token")
    del_by_path(input_dict, "context", "eip712", "domain", "version")

    # Adapt formats objects to match EIP-712 conversion
    def _cleanup_formats(formats: dict[str, Any]) -> Any:
        for _, message in formats.items():
            # Remove ERC-7730 specific fields
            del_by_path(message, "$id")
            del_by_path(message, "required")
            del_by_path(message, "excluded")
            del_by_path(message, "screens")
            if "fields" in message:
                for field in message["fields"]:
                    if "format" in field:
                        format = field["format"]

                        # Address name format parameters cannot be preserved
                        if format == "addressName":
                            del_by_path(field, "params")

                        # Other formats are always converted to RAW
                        if format not in (
                            FieldFormat.AMOUNT,
                            FieldFormat.TOKEN_AMOUNT,
                            FieldFormat.DATE,
                        ):
                            field["format"] = "raw"

        return formats

    map_by_path(input_dict, _cleanup_formats, "display", "formats")

    assert_dict_equals(input_dict, output_dict)


def _assert_eip712_json_equals_with_tolerance(
    input: InputEIP712DAppDescriptor, output: InputEIP712DAppDescriptor
) -> None:
    input_dict, output_dict = json.loads(model_to_json_str(input)), json.loads(model_to_json_str(output))

    # format is required in ERC-7730 schema, and missing format in EIP-712 is equivalent to raw
    def cleanup_dapp_descriptor(dapp: dict[str, Any]) -> Any:
        for contract in dapp["contracts"]:
            for message in contract["messages"]:
                for field in message["mapper"]["fields"]:
                    if field.get("format") == "raw":
                        field.pop("format", None)

    cleanup_dapp_descriptor(input_dict)
    cleanup_dapp_descriptor(output_dict)

    assert_dict_equals(input_dict, output_dict)



from abc import ABC, abstractmethod
from enum import StrEnum, auto
from typing import Generic, TypeVar

from erc7730.model.abi import ABI
from erc7730.model.context import EIP712Schema


class TxClass(StrEnum):
    STAKE = auto()
    SWAP = auto()
    PERMIT = auto()
    WITHDRAW = auto()


Schema = TypeVar("Schema", list[ABI], EIP712Schema)


class Classifier(ABC, Generic[Schema]):
    """Given a schema (which is an EIP712 schema or an ABI schema), classify the transaction type
    with some predefined ruleset.
    """

    @abstractmethod
    def classify(self, schema: Schema) -> TxClass | None:
        raise NotImplementedError()



from pydantic import Field
from pydantic_string_url import HttpUrl

from erc7730.model.abi import ABI
from erc7730.model.base import Model
from erc7730.model.context import EIP712Schema
from erc7730.model.types import Address, Id

# ruff: noqa: N815 - camel case field names are tolerated to match schema


class ResolvedDomain(Model):
    """
    EIP 712 Domain Binding constraint.

    Each value of the domain constraint MUST match the corresponding eip 712 message domain value.
    """

    name: str | None = Field(default=None, title="Name", description="The EIP-712 domain name.")

    version: str | None = Field(default=None, title="Version", description="The EIP-712 version.")

    chainId: int | None = Field(default=None, title="Chain ID", description="The EIP-155 chain id.")

    verifyingContract: Address | None = Field(
        default=None, title="Verifying Contract", description="The EIP-712 verifying contract address."
    )


class ResolvedDeployment(Model):
    """
    A deployment describing where the contract is deployed.

    The target contract (Tx to or factory) MUST match one of those deployments.
    """

    chainId: int = Field(title="Chain ID", description="The deployment EIP-155 chain id.")

    address: Address = Field(title="Contract Address", description="The deployment contract address.")


class ResolvedFactory(Model):
    """
    A factory constraint is used to check whether the target contract is deployed by a specified factory.
    """

    deployments: list[ResolvedDeployment] = Field(
        title="Deployments",
        description="An array of deployments describing where the contract is deployed. The target contract (Tx to or"
        "factory) MUST match one of those deployments.",
    )

    deployEvent: str = Field(
        title="Deploy Event signature",
        description="The event signature that the factory emits when deploying a new contract.",
    )


class ResolvedBindingContext(Model):
    deployments: list[ResolvedDeployment] = Field(
        title="Deployments",
        description="An array of deployments describing where the contract is deployed. The target contract (Tx to or"
        "factory) MUST match one of those deployments.",
        min_length=1,
    )


class ResolvedContract(ResolvedBindingContext):
    """
    Contract Binding Context.

    The contract binding context is a set constraints that are used to bind the ERC7730 file to a specific smart
    contract.
    """

    abi: list[ABI] = Field(
        title="ABI",
        description="The ABI of the target contract. This must an array of ABI objects.",
    )

    addressMatcher: HttpUrl | None = Field(
        default=None,
        title="Address Matcher",
        description="An URL of a contract address matcher that should be used to match the contract address.",
    )

    factory: ResolvedFactory | None = Field(
        default=None,
        title="Factory Constraint",
        description="A factory constraint is used to check whether the target contract is deployed by a specified"
        "factory.",
    )


class ResolvedEIP712(ResolvedBindingContext):
    """
    EIP 712 Binding.

    The EIP-712 binding context is a set of constraints that must be verified by the message being signed.
    """

    domain: ResolvedDomain | None = Field(
        default=None,
        title="EIP 712 Domain Binding constraint",
        description="Each value of the domain constraint MUST match the corresponding eip 712 message domain value.",
    )

    domainSeparator: str | None = Field(
        default=None,
        title="Domain Separator constraint",
        description="The domain separator value that must be matched by the message. In hex string representation.",
    )

    schemas: list[EIP712Schema] = Field(title="EIP-712 messages schemas", description="Schemas of all messages.")


class ResolvedContractContext(Model):
    """
    Contract Binding Context.

    The contract binding context is a set constraints that are used to bind the ERC7730 file to a specific smart
    contract.
    """

    id: Id | None = Field(
        alias="$id",
        default=None,
        title="Id",
        description="An internal identifier that can be used either for clarity specifying what the element is or as a"
        "reference in device specific sections.",
    )

    contract: ResolvedContract = Field(
        title="Contract Binding Context",
        description="The contract binding context is a set constraints that are used to bind the ERC7730 file to a"
        "specific smart contract.",
    )


class ResolvedEIP712Context(Model):
    """
    EIP 712 Binding.

    The EIP-712 binding context is a set of constraints that must be verified by the message being signed.
    """

    id: Id | None = Field(
        alias="$id",
        default=None,
        title="Id",
        description="An internal identifier that can be used either for clarity specifying what the element is or as a"
        "reference in device specific sections.",
    )

    eip712: ResolvedEIP712 = Field(
        title="EIP 712 Binding",
        description="The EIP-712 binding context is a set of constraints that must be verified by the message being"
        "signed.",
    )



"""
Object model for ERC-7730 descriptors `metadata` section.

Specification: https://github.com/LedgerHQ/clear-signing-erc7730-registry/tree/master/specs
JSON schema: https://github.com/LedgerHQ/clear-signing-erc7730-registry/blob/master/specs/erc7730-v1.schema.json
"""

from datetime import UTC, datetime
from typing import Annotated

from pydantic import Field
from pydantic_string_url import HttpUrl

from erc7730.model.base import Model

# ruff: noqa: N815 - camel case field names are tolerated to match schema


class OwnerInfo(Model):
    """
    Main contract's owner detailed information.

    The owner info section contains detailed information about the owner or target of the contract / message to be
    clear signed.
    """

    legalName: str = Field(
        title="Owner Legal Name",
        description="The full legal name of the owner if different from the owner field.",
        min_length=1,
        examples=["Tether Limited", "Lido DAO"],
    )

    lastUpdate: datetime | None = Field(
        default=None,
        title="[DEPRECATED] Last Update of the contract / message",
        description="The date of the last update of the contract / message."
        "Field is deprecated and will be removed, use `deploymentDate` instead.",
        examples=[datetime.now(UTC)],
    )

    deploymentDate: datetime | None = Field(
        default=None,
        title="Deployment date of contract / message",
        description="The deployment date of the contract / message.",
        examples=[datetime.now(UTC)],
    )

    url: HttpUrl = Field(
        title="Owner URL",
        description="URL with more info on the entity the user interacts with.",
        examples=[HttpUrl("https://tether.to"), HttpUrl("https://lido.fi")],
    )


class TokenInfo(Model):
    """
    Token Description.

    A description of an ERC20 token exported by this format, that should be trusted. Not mandatory if the
    corresponding metadata can be fetched from the contract itself.
    """

    name: str = Field(
        title="Token Name",
        description="The token display name.",
        min_length=1,
        max_length=255,  # TODO: arbitrary value, to be refined
        examples=["Tether USD", "Dai Stablecoin"],
    )

    ticker: str = Field(
        title="Token Ticker",
        description="A short capitalized ticker for the token, that will be displayed in front of corresponding "
        "amounts.",
        min_length=1,
        max_length=10,  # TODO: arbitrary value, to be refined
        pattern=r"^[a-zA-Z0-9_\\-\\.]+$",
        examples=["USDT", "DAI", "rsETH"],
    )

    decimals: int = Field(
        title="Token Decimals",
        description="The number of decimals of the token ticker, used to display amounts.",
        ge=0,
        le=255,
        examples=[0, 18],
    )


class Metadata(Model):
    """
    Metadata Section.

    The metadata section contains information about constant values relevant in the scope of the current contract /
    message (as matched by the `context` section)
    """

    owner: str | None = Field(
        default=None,
        title="Owner display name.",
        description="The display name of the owner or target of the contract / message to be clear signed.",
    )

    info: OwnerInfo | None = Field(
        default=None,
        title="Main contract's owner detailed information.",
        description="The owner info section contains detailed information about the owner or target of the contract / "
        "message to be clear signed.",
    )

    token: TokenInfo | None = Field(
        default=None,
        title="Token Description",
        description="A description of an ERC20 token exported by this format, that should be trusted. Not mandatory if "
        "the corresponding metadata can be fetched from the contract itself.",
    )


EnumDefinition = Annotated[
    dict[str, str],
    Field(
        title="Enum Definition",
        description="A mapping of enum values to human readable strings.",
        examples=[{"1": "stable", "2": "variable"}],
        min_length=1,
        max_length=32,
    ),
]



from typing import Any

from lark import Lark, UnexpectedInput
from lark.exceptions import VisitError
from lark.visitors import Transformer_InPlaceRecursive
from pydantic import TypeAdapter, ValidationError

from erc7730.model.paths import (
    Array,
    ArrayElement,
    ArrayIndex,
    ArraySlice,
    ContainerField,
    ContainerPath,
    DataPath,
    DescriptorPath,
    Field,
)

PATH_PARSER = Lark(
    grammar=r"""
        ?path: descriptor_path | container_path | data_path
    
        descriptor_path: "$." descriptor_path_component ("." descriptor_path_component)*
        ?descriptor_path_component: field | array_element
    
        container_path: "@." container_field
        !container_field: "from" | "to" | "value"
    
        ?data_path: absolute_data_path | relative_data_path
        absolute_data_path: "#." data_path_component ("." data_path_component)*
        relative_data_path: data_path_component ("." data_path_component)*
        ?data_path_component: field | array | array_element | array_slice

        field: /[a-zA-Z0-9_]+/
        array: "[]"
        array_index: /-?[0-9]+/
        array_element: "[" array_index "]"
        slice_array_index: array_index?
        array_slice: "[" slice_array_index ":" slice_array_index "]"
    """,
    start="path",
)


class PathTransformer(Transformer_InPlaceRecursive):
    """Visitor to transform the parsed path AST into path domain model objects."""

    def field(self, ast: Any) -> Field:
        (value,) = ast
        return Field(identifier=value.value)

    def array(self, ast: Any) -> Array:
        return Array()

    def array_index(self, ast: Any) -> ArrayIndex:
        (value,) = ast
        return TypeAdapter(ArrayIndex).validate_strings(value)

    def array_element(self, ast: Any) -> ArrayElement:
        (value,) = ast
        return ArrayElement(index=value)

    def slice_array_index(self, ast: Any) -> ArrayIndex | None:
        if len(ast) == 1:
            (value,) = ast
            return value
        return None

    def array_slice(self, ast: Any) -> ArraySlice:
        (start, end) = ast
        return ArraySlice(start=start, end=end)

    def container_field(self, ast: Any) -> ContainerField:
        (value,) = ast
        return ContainerField(value)

    def descriptor_path(self, ast: Any) -> DescriptorPath:
        return DescriptorPath(elements=ast)

    def container_path(self, ast: Any) -> ContainerPath:
        (value,) = ast
        return ContainerPath(field=value)

    def absolute_data_path(self, ast: Any) -> DataPath:
        return DataPath(elements=ast, absolute=True)

    def relative_data_path(self, ast: Any) -> DataPath:
        return DataPath(elements=ast, absolute=False)


PATH_TRANSFORMER = PathTransformer()


def to_path(path: str) -> ContainerPath | DataPath | DescriptorPath:
    """
    Parse a path string into a domain model object.

    :param path: the path input string
    :return: an union of all possible path types
    :raises ValueError: if the input string is not a valid path
    :raises Exception: if the path parsing fails for an unexpected reason
    """
    try:
        return PATH_TRANSFORMER.transform(PATH_PARSER.parse(path))
    except UnexpectedInput as e:
        # TODO improve error reporting, see:
        #  https://github.com/lark-parser/lark/blob/master/examples/advanced/error_reporting_lalr.py
        raise ValueError(f"""Invalid path "{path}": {e}""") from None
    except VisitError as e:
        if isinstance(e.orig_exc, ValidationError):
            raise ValueError(f"""Invalid path "{path}": {e.orig_exc}`""") from None
        raise Exception(
            f"""Failed to parse path "{path}": {e}`\n"""
            "This is most likely a bug in the ERC-7730 library, please report it to authors."
        ) from e
    except Exception as e:
        raise Exception(
            f"""Failed to parse path "{path}": {e}`\n"""
            "This is most likely a bug in the ERC-7730 library, please report it to authors."
        ) from e



"""
Base model for library, using pydantic.

See https://docs.pydantic.dev
"""

from pathlib import Path
from typing import Self

from pydantic import BaseModel, ConfigDict

from erc7730.common.pydantic import (
    model_from_json_file_with_includes,
    model_from_json_file_with_includes_or_none,
    model_to_json_file,
    model_to_json_str,
)


class Model(BaseModel):
    """
    Base model for library, using pydantic.

    See https://docs.pydantic.dev
    """

    model_config = ConfigDict(
        strict=True,
        frozen=True,
        extra="forbid",
        validate_default=True,
        validate_return=True,
        validate_assignment=True,
        arbitrary_types_allowed=False,
        allow_inf_nan=False,
    )

    @classmethod
    def load(cls, path: Path) -> Self:
        """
        Load a model from a JSON file.

        :param path: file path
        :return: validated in-memory representation of model
        :raises Exception: if the file does not exist or has validation errors
        """
        return model_from_json_file_with_includes(path, cls)

    @classmethod
    def load_or_none(cls, path: Path) -> Self | None:
        """
        Load a model from a JSON file.

        :param path: file path
        :return: validated in-memory representation of descriptor, or None if file does not exist
        :raises Exception: if the file has validation errors
        """
        return model_from_json_file_with_includes_or_none(path, cls)

    def save(self, path: Path) -> None:
        """
        Write a model to a JSON file, creating parent directories as needed.
        """
        model_to_json_file(path, self)

    def to_json_string(self) -> str:
        """
        Serialize the model to a JSON string.

        :return: JSON representation of model, serialized as a string
        """
        return model_to_json_str(self)



"""
Package implementing an object model for ERC-7730 resolved descriptors.

This model represents descriptors after resolution phase:
    - URLs have been fetched
    - Contract addresses have been normalized to lowercase
    - References have been inlined
    - Constants have been inlined
    - Field definitions have been inlined
    - Nested fields have been flattened where possible
    - Selectors have been converted to 4 bytes form
"""



from collections.abc import Generator
from typing import Any, assert_never

from caseswitcher import to_title
from pydantic import TypeAdapter
from pydantic_string_url import HttpUrl

from erc7730.common.abi import ABIDataType, compute_signature, get_functions
from erc7730.common.client import get_contract_abis
from erc7730.generate.schema_tree import (
    SchemaArray,
    SchemaLeaf,
    SchemaStruct,
    SchemaTree,
    abi_function_to_tree,
    eip712_schema_to_tree,
)
from erc7730.model.abi import ABI
from erc7730.model.context import EIP712Schema
from erc7730.model.display import AddressNameType, DateEncoding, FieldFormat
from erc7730.model.input.context import (
    InputContract,
    InputContractContext,
    InputDeployment,
    InputEIP712,
    InputEIP712Context,
)
from erc7730.model.input.descriptor import InputERC7730Descriptor
from erc7730.model.input.display import (
    InputAddressNameParameters,
    InputDateParameters,
    InputDisplay,
    InputField,
    InputFieldDescription,
    InputFieldParameters,
    InputFormat,
    InputNestedFields,
)
from erc7730.model.input.metadata import InputMetadata
from erc7730.model.metadata import OwnerInfo
from erc7730.model.paths import ROOT_DATA_PATH, Array, ArrayElement, ArraySlice, DataPath, Field
from erc7730.model.paths.path_ops import data_path_append
from erc7730.model.types import Address


def generate_descriptor(
    chain_id: int,
    contract_address: Address,
    abi: str | bytes | None = None,
    eip712_schema: str | bytes | None = None,
    owner: str | None = None,
    legal_name: str | None = None,
    url: HttpUrl | None = None,
) -> InputERC7730Descriptor:
    """
    Generate an ERC-7730 descriptor.

    If an EIP-712 schema is provided, an EIP-712 descriptor is generated for this schema, otherwise a calldata
    descriptor. If no ABI is supplied, the ABIs are fetched from Etherscan using the chain id / contract address.

    :param chain_id: contract chain id
    :param contract_address: contract address
    :param abi: JSON ABI string or buffer representation (to generate a calldata descriptor)
    :param eip712_schema: JSON EIP-712 schema string or buffer representation (to generate an EIP-712 descriptor)
    :param owner: the display name of the owner or target of the contract / message to be clear signed
    :param legal_name: the full legal name of the owner if different from the owner field
    :param url: URL with more info on the entity the user interacts with
    :return: a generated ERC-7730 descriptor
    """

    context, trees = _generate_context(chain_id, contract_address, abi, eip712_schema)
    metadata = _generate_metadata(legal_name, owner, url)
    display = _generate_display(trees)

    return InputERC7730Descriptor(context=context, metadata=metadata, display=display)


def _generate_metadata(owner: str | None, legal_name: str | None, url: HttpUrl | None) -> InputMetadata:
    info = OwnerInfo(legalName=legal_name, url=url) if legal_name is not None and url is not None else None
    return InputMetadata(owner=owner, info=info)


def _generate_context(
    chain_id: int, contract_address: Address, abi: str | bytes | None, eip712_schema: str | bytes | None
) -> tuple[InputContractContext | InputEIP712Context, dict[str, SchemaTree]]:
    if eip712_schema is not None:
        return _generate_context_eip712(chain_id, contract_address, eip712_schema)
    return _generate_context_calldata(chain_id, contract_address, abi)


def _generate_context_eip712(
    chain_id: int, contract_address: Address, eip712_schema: str | bytes
) -> tuple[InputEIP712Context, dict[str, SchemaTree]]:
    schemas = TypeAdapter(list[EIP712Schema]).validate_json(eip712_schema)

    context = InputEIP712Context(
        eip712=InputEIP712(schemas=schemas, deployments=[InputDeployment(chainId=chain_id, address=contract_address)])
    )

    trees = {schema.primaryType: eip712_schema_to_tree(schema) for schema in schemas}

    return context, trees


def _generate_context_calldata(
    chain_id: int, contract_address: Address, abi: str | bytes | None
) -> tuple[InputContractContext, dict[str, SchemaTree]]:
    if abi is not None:
        abis = TypeAdapter(list[ABI]).validate_json(abi)

    elif (abis := get_contract_abis(chain_id, contract_address)) is None:
        raise Exception("Failed to fetch contract ABIs")

    functions = list(get_functions(abis).functions.values())

    context = InputContractContext(
        contract=InputContract(abi=functions, deployments=[InputDeployment(chainId=chain_id, address=contract_address)])
    )

    trees = {compute_signature(function): abi_function_to_tree(function) for function in functions}

    return context, trees


def _generate_display(trees: dict[str, SchemaTree]) -> InputDisplay:
    return InputDisplay(formats=_generate_formats(trees))


def _generate_formats(trees: dict[str, SchemaTree]) -> dict[str, InputFormat]:
    formats: dict[str, InputFormat] = {}
    for name, tree in trees.items():
        if fields := list(_generate_fields(schema=tree, path=ROOT_DATA_PATH)):
            formats[name] = InputFormat(fields=fields)
    return formats


def _generate_fields(schema: SchemaTree, path: DataPath) -> Generator[InputField, Any, Any]:
    match schema:
        case SchemaStruct(components=components) if path == ROOT_DATA_PATH:
            for name, component in components.items():
                if name:
                    yield from _generate_fields(component, data_path_append(path, Field(identifier=name)))

        case SchemaStruct(components=components):
            fields = [
                field
                for name, component in components.items()
                for field in _generate_fields(component, DataPath(absolute=False, elements=[Field(identifier=name)]))
                if name
            ]
            yield InputNestedFields(path=path, fields=fields)

        case SchemaArray(component=component):
            match component:
                case SchemaStruct() | SchemaArray():
                    yield InputNestedFields(
                        path=data_path_append(path, Array()),
                        fields=list(_generate_fields(component, DataPath(absolute=False, elements=[]))),
                    )
                case SchemaLeaf():
                    yield from _generate_fields(component, data_path_append(path, Array()))
                case _:
                    assert_never(schema)

        case SchemaLeaf(data_type=data_type):
            name = _get_leaf_name(path)
            format, params = _generate_field(name, data_type)
            yield InputFieldDescription(path=path, label=name, format=format, params=params)

        case _:
            assert_never(schema)


def _generate_field(name: str, data_type: ABIDataType) -> tuple[FieldFormat, InputFieldParameters | None]:
    match data_type:
        case ABIDataType.UINT | ABIDataType.INT:
            # other applicable formats could be TOKEN_AMOUNT, UNIT or ENUM, but we can't tell

            if _contains_any_of(name, "duration"):
                return FieldFormat.DURATION, None

            if _contains_any_of(name, "height"):
                return FieldFormat.DATE, InputDateParameters(encoding=DateEncoding.BLOCKHEIGHT)

            if _contains_any_of(name, "deadline", "expiration", "until", "time", "timestamp"):
                return FieldFormat.DATE, InputDateParameters(encoding=DateEncoding.TIMESTAMP)

            if _contains_any_of(name, "amount", "value", "price"):
                return FieldFormat.AMOUNT, None

            return FieldFormat.RAW, None

        case ABIDataType.UFIXED | ABIDataType.FIXED:
            return FieldFormat.RAW, None

        case ABIDataType.ADDRESS:
            if _contains_any_of(name, "collection", "nft"):
                return FieldFormat.NFT_NAME, InputAddressNameParameters(types=[AddressNameType.COLLECTION])

            if _contains_any_of(name, "spender"):
                return FieldFormat.ADDRESS_NAME, InputAddressNameParameters(types=[AddressNameType.CONTRACT])

            if _contains_any_of(name, "asset", "token"):
                return FieldFormat.ADDRESS_NAME, InputAddressNameParameters(types=[AddressNameType.TOKEN])

            if _contains_any_of(name, "from", "to", "owner", "recipient", "receiver", "account"):
                return FieldFormat.ADDRESS_NAME, InputAddressNameParameters(
                    types=[AddressNameType.EOA, AddressNameType.WALLET]
                )

            return FieldFormat.ADDRESS_NAME, InputAddressNameParameters(types=list(AddressNameType))

        case ABIDataType.BOOL:
            return FieldFormat.RAW, None

        case ABIDataType.BYTES:
            if _contains_any_of(name, "calldata"):
                return FieldFormat.CALL_DATA, None

            return FieldFormat.RAW, None

        case ABIDataType.STRING:
            return FieldFormat.RAW, None

        case _:
            assert_never(data_type)


def _get_leaf_name(path: DataPath) -> str:
    for element in reversed(path.elements):
        match element:
            case Field(identifier=name):
                return to_title(name).strip()
            case Array() | ArrayElement() | ArraySlice():
                continue
            case _:
                assert_never(element)
    return "unknown"


def _contains_any_of(name: str, *values: str) -> bool:
    name_lower = name.lower()
    return any(value in name_lower for value in values)



"""
Object model for ERC-7730 descriptors `metadata` section.

Specification: https://github.com/LedgerHQ/clear-signing-erc7730-registry/tree/master/specs
JSON schema: https://github.com/LedgerHQ/clear-signing-erc7730-registry/blob/master/specs/erc7730-v1.schema.json
"""

from pydantic import Field

from erc7730.model.metadata import EnumDefinition, Metadata
from erc7730.model.types import Id

# ruff: noqa: N815 - camel case field names are tolerated to match schema


class ResolvedMetadata(Metadata):
    """
    Metadata Section.

    The metadata section contains information about constant values relevant in the scope of the current contract /
    message (as matched by the `context` section)
    """

    enums: dict[Id, EnumDefinition] | None = Field(
        default=None,
        title="Enums",
        description="A set of enums that are used to format fields replacing values with human readable strings.",
        examples=[{"interestRateMode": {"1": "stable", "2": "variable"}}],
        max_length=32,  # TODO refine
    )



from typing import Annotated

from pydantic import Field as PydanticField
from pydantic import GetPydanticSchema
from pydantic_core import core_schema
from pydantic_core.core_schema import (
    chain_schema,
    is_instance_schema,
    json_or_python_schema,
    no_info_plain_validator_function,
    str_schema,
    to_string_ser_schema,
)

from erc7730.common.pydantic import ErrorTypeLabel
from erc7730.model.paths import ContainerPath, DataPath, DescriptorPath
from erc7730.model.paths.path_parser import to_path

CONTAINER_PATH_STR_JSON_SCHEMA = chain_schema(
    [str_schema(), no_info_plain_validator_function(to_path), is_instance_schema(ContainerPath)]
)
CONTAINER_PATH_STR_CORE_SCHEMA = json_or_python_schema(
    json_schema=CONTAINER_PATH_STR_JSON_SCHEMA,
    python_schema=core_schema.union_schema([is_instance_schema(ContainerPath), CONTAINER_PATH_STR_JSON_SCHEMA]),
    serialization=to_string_ser_schema(),
)
ContainerPathStr = Annotated[
    ContainerPath,
    GetPydanticSchema(lambda _type, _handler: CONTAINER_PATH_STR_CORE_SCHEMA),
    PydanticField(
        title="Input Path",
        description="A path applying to the container of the structured data to be signed. Such paths are prefixed "
        """with "@".""",
    ),
    ErrorTypeLabel(
        "path applying to the container of the structured data to be signed, such as " + '"@.to" or "@.value".'
    ),
]

DATA_PATH_STR_JSON_SCHEMA = chain_schema(
    [str_schema(), no_info_plain_validator_function(to_path), is_instance_schema(DataPath)]
)
DATA_PATH_STR_CORE_SCHEMA = json_or_python_schema(
    json_schema=DATA_PATH_STR_JSON_SCHEMA,
    python_schema=core_schema.union_schema([is_instance_schema(DataPath), DATA_PATH_STR_JSON_SCHEMA]),
    serialization=to_string_ser_schema(),
)
DataPathStr = Annotated[
    DataPath,
    GetPydanticSchema(lambda _type, _handler: DATA_PATH_STR_CORE_SCHEMA),
    PydanticField(
        title="Data Path",
        description="A path applying to the structured data schema (ABI path for contracts, path in the message types "
        "itself for EIP-712). A data path can reference multiple values if it contains array elements or slices. Such "
        """paths are prefixed with "#".""",
    ),
    ErrorTypeLabel("data path referencing a value in the structured data schema, such as " + '"#.foo.[0].bar.[0:-1]".'),
]

DESCRIPTOR_PATH_STR_JSON_SCHEMA = chain_schema(
    [str_schema(), no_info_plain_validator_function(to_path), is_instance_schema(DescriptorPath)]
)
DESCRIPTOR_PATH_STR_CORE_SCHEMA = json_or_python_schema(
    json_schema=DESCRIPTOR_PATH_STR_JSON_SCHEMA,
    python_schema=core_schema.union_schema([is_instance_schema(DescriptorPath), DESCRIPTOR_PATH_STR_JSON_SCHEMA]),
    serialization=to_string_ser_schema(),
)
DescriptorPathStr = Annotated[
    DescriptorPath,
    GetPydanticSchema(lambda _type, _handler: DESCRIPTOR_PATH_STR_CORE_SCHEMA),
    PydanticField(
        title="Descriptor Path",
        description="A path applying to the current file describing the structured data formatting, after merging "
        "with includes. A descriptor path can only reference a single value in the document. Such paths are prefixed "
        """with "$".""",
        examples=["$.foo.bar"],
    ),
    ErrorTypeLabel(
        "descriptor path referencing a value in the descriptor document using jsonpath syntax, such as "
        + '"$.foo.bar".'
    ),
]



import json
import os
from abc import ABC
from functools import cache
from typing import Any, TypeVar, final, override

from hishel import CacheTransport, FileStorage
from httpx import URL, BaseTransport, Client, HTTPTransport, Request, Response
from httpx._content import IteratorByteStream
from httpx_file import FileTransport
from limiter import Limiter
from pydantic import ConfigDict, TypeAdapter, ValidationError
from pydantic_string_url import FileUrl, HttpUrl
from xdg_base_dirs import xdg_cache_home

from erc7730.model.abi import ABI
from erc7730.model.base import Model
from erc7730.model.types import Address

ETHERSCAN = "api.etherscan.io"

_T = TypeVar("_T")


class EtherscanChain(Model):
    """Etherscan supported chain info."""

    model_config = ConfigDict(strict=False, frozen=True, extra="ignore")
    chainname: str
    chainid: int
    blockexplorer: HttpUrl


@cache
def get_supported_chains() -> list[EtherscanChain]:
    """
    Get supported chains from Etherscan.

    :return: Etherscan supported chains, with name/chain id/block explorer URL
    """
    return get(url=HttpUrl(f"https://{ETHERSCAN}/v2/chainlist"), model=list[EtherscanChain])


def get_contract_abis(chain_id: int, contract_address: Address) -> list[ABI]:
    """
    Get contract ABIs from Etherscan.

    :param chain_id: EIP-155 chain ID
    :param contract_address: EVM contract address
    :return: deserialized list of ABIs
    :raises NotImplementedError: if chain id not supported, API key not setup, or unexpected response
    """
    try:
        return get(
            url=HttpUrl(f"https://{ETHERSCAN}/v2/api"),
            chainid=chain_id,
            module="contract",
            action="getabi",
            address=contract_address,
            model=list[ABI],
        )
    except Exception as e:
        if "Contract source code not verified" in str(e):
            raise Exception("contract source is not available on Etherscan") from e
        if "Max calls per sec rate limit reached" in str(e):
            raise Exception("Etherscan rate limit exceeded, please retry") from e
        raise e


def get_contract_explorer_url(chain_id: int, contract_address: Address) -> HttpUrl:
    """
    Get contract explorer site URL (for opening in a browser).

    :param chain_id: EIP-155 chain ID
    :param contract_address: EVM contract address
    :return: URL to the contract explorer site
    :raises NotImplementedError: if chain id not supported
    """
    for chain in get_supported_chains():
        if chain.chainid == chain_id:
            return HttpUrl(f"{chain.blockexplorer}/address/{contract_address}#code")
    raise NotImplementedError(
        f"Chain ID {chain_id} is not supported, please report this to authors of python-erc7730 library"
    )


def get(model: type[_T], url: HttpUrl | FileUrl, **params: Any) -> _T:
    """
    Fetch data from a file or an HTTP URL and deserialize it.

    This method implements some automated adaptations to handle user provided URLs:
     - GitHub: adaptation to "raw.githubusercontent.com"
     - Etherscan: rate limiting, API key parameter injection, "result" field unwrapping

    :param url: URL to get data from
    :param model: Pydantic model to deserialize the data
    :return: deserialized response
    :raises Exception: if URL type is not supported, API key not setup, or unexpected response
    """
    with _client() as client:
        response = client.get(url, params=params).raise_for_status().content
    try:
        return TypeAdapter(model).validate_json(response)
    except ValidationError as e:
        raise Exception(f"Received unexpected response from {url}: {response.decode(errors='replace')}") from e


def _client() -> Client:
    """
    Create a new HTTP client with GitHub and Etherscan specific transports.
    :return:
    """
    cache_storage = FileStorage(base_path=xdg_cache_home() / "erc7730", ttl=7 * 24 * 3600, check_ttl_every=24 * 3600)
    http_transport = HTTPTransport()
    http_transport = GithubTransport(http_transport)
    http_transport = EtherscanTransport(http_transport)
    http_transport = CacheTransport(transport=http_transport, storage=cache_storage)
    file_transport = FileTransport()
    # TODO file storage: authorize relative paths only
    transports = {"https://": http_transport, "file://": file_transport}
    return Client(mounts=transports, timeout=10)


class DelegateTransport(ABC, BaseTransport):
    """Base class for wrapping httpx transport."""

    def __init__(self, delegate: BaseTransport) -> None:
        self._delegate = delegate

    def handle_request(self, request: Request) -> Response:
        return self._delegate.handle_request(request)

    def close(self) -> None:
        self._delegate.close()


@final
class GithubTransport(DelegateTransport):
    """GitHub specific transport for handling raw content requests."""

    GITHUB, GITHUB_RAW = "github.com", "raw.githubusercontent.com"

    def __init__(self, delegate: BaseTransport) -> None:
        super().__init__(delegate)

    @override
    def handle_request(self, request: Request) -> Response:
        if request.url.host != self.GITHUB:
            return super().handle_request(request)

        # adapt URL
        request.url = URL(str(request.url).replace(self.GITHUB, self.GITHUB_RAW).replace("/blob/", "/"))
        request.headers.update({"Host": self.GITHUB_RAW})
        return super().handle_request(request)


@final
class EtherscanTransport(DelegateTransport):
    """Etherscan specific transport for handling rate limiting, API key parameter injection, response unwrapping."""

    ETHERSCAN_API_HOST = "ETHERSCAN_API_HOST"
    ETHERSCAN_API_KEY = "ETHERSCAN_API_KEY"

    @Limiter(rate=5, capacity=5, consume=1)
    @override
    def handle_request(self, request: Request) -> Response:
        if request.url.host != ETHERSCAN:
            return super().handle_request(request)

        # substitute base URL if provided
        if (api_host := os.environ.get(self.ETHERSCAN_API_HOST)) is not None:
            request.url = request.url.copy_with(host=api_host)
            request.headers.update({"Host": api_host})

        # add API key if provided
        if (api_key := os.environ.get(self.ETHERSCAN_API_KEY)) is not None or (
            api_key := os.environ.get(f"SCAN_{self.ETHERSCAN_API_KEY}")
        ) is not None:
            request.url = request.url.copy_add_param("apikey", api_key)

        # read response
        response = super().handle_request(request)
        response.read()
        response.close()

        # unwrap result, sometimes containing JSON directly, sometimes JSON in a string
        try:
            if (result := response.json().get("result")) is not None:
                data = result if isinstance(result, str) else json.dumps(result)
                return Response(status_code=response.status_code, stream=IteratorByteStream([data.encode()]))
        except Exception:
            pass  # nosec B110 - intentional try/except/pass

        raise Exception(f"Unexpected response from Etherscan: {response.content.decode(errors='replace')}")



from glob import glob
from pathlib import Path

import pytest
from rich import print

from erc7730.generate.generate import generate_descriptor
from erc7730.model.input.descriptor import InputERC7730Descriptor
from erc7730.model.input.lenses import get_deployments
from erc7730.model.types import Address

DATA = Path(__file__).resolve().parent / "data"


@pytest.mark.parametrize(
    "label,chain_id,contract_address",
    [
        ("Tether USD", 42161, Address("0xfd086bc7cd5c481dcc9c85ebe478a1c0b69fcbb9")),
        ("POAP Bridge", 1, Address("0x0bb4d3e88243f4a057db77341e6916b0e449b158")),
        ("QuickSwap", 137, Address("0xa5E0829CaCEd8fFDD4De3c43696c57F7D7A678ff")),
        ("1inch Aggregation router v5", 10, Address("0x1111111254EEB25477B68fb85Ed929f73A960582")),
        ("Paraswap v5 AugustusSwapper", 137, Address("0xDEF171Fe48CF0115B1d80b88dc8eAB59176FEe57")),
        ("Uniswap v3 Router 2", 8453, Address("0x3fC91A3afd70395Cd496C647d5a6CC9D4B2b7FAD")),
    ],
)
def test_generate_from_contract_address(label: str, chain_id: int, contract_address: Address) -> None:
    """Generate descriptor using a provided ABI file."""
    _assert_descriptor_valid(generate_descriptor(chain_id=chain_id, contract_address=contract_address, owner=label))


@pytest.mark.parametrize("test_file", sorted(glob(str(DATA / "abis*.json"))), ids=lambda f: Path(f).stem)
def test_generate_from_abis_buffer(test_file: str) -> None:
    """Generate descriptor using a provided ABI file."""
    with open(test_file, "rb") as f:
        _assert_descriptor_valid(
            generate_descriptor(
                chain_id=1, contract_address=Address("0x0000000000000000000000000000000000000000"), abi=f.read()
            )
        )


@pytest.mark.parametrize("test_file", sorted(glob(str(DATA / "abis*.json"))), ids=lambda f: Path(f).stem)
def test_generate_from_abis_string(test_file: str) -> None:
    """Generate descriptor using a provided ABI file."""
    with open(test_file, "rb") as f:
        abi = f.read().decode("utf-8")
    _assert_descriptor_valid(
        generate_descriptor(chain_id=1, contract_address=Address("0x0000000000000000000000000000000000000000"), abi=abi)
    )


@pytest.mark.parametrize("test_file", sorted(glob(str(DATA / "schemas*.json"))), ids=lambda f: Path(f).stem)
def test_generate_from_eip712_schemas_buffer(test_file: str) -> None:
    """Generate descriptor using a provided EIP-712 file."""
    with open(test_file, "rb") as f:
        _assert_descriptor_valid(
            generate_descriptor(
                chain_id=1,
                contract_address=Address("0x0000000000000000000000000000000000000000"),
                eip712_schema=f.read(),
            )
        )


@pytest.mark.parametrize("test_file", sorted(glob(str(DATA / "schemas*.json"))), ids=lambda f: Path(f).stem)
def test_generate_from_eip712_schemas_string(test_file: str) -> None:
    """Generate descriptor using a provided EIP-712 file."""
    with open(test_file, "rb") as f:
        schema = f.read().decode("utf-8")
    _assert_descriptor_valid(
        generate_descriptor(
            chain_id=1,
            contract_address=Address("0x0000000000000000000000000000000000000000"),
            eip712_schema=schema,
        )
    )


def _assert_descriptor_valid(descriptor: InputERC7730Descriptor) -> None:
    print(descriptor.to_json_string())
    assert len(get_deployments(descriptor)) == 1
    assert len(descriptor.display.formats) > 0
    for format in descriptor.display.formats.values():
        assert len(format.fields) > 0



from typing import final, override

from erc7730.common.output import OutputAdder
from erc7730.lint import ERC7730Linter
from erc7730.lint.classifier import TxClass
from erc7730.lint.classifier.abi_classifier import ABIClassifier
from erc7730.lint.classifier.eip712_classifier import EIP712Classifier
from erc7730.model.resolved.context import EIP712Schema, ResolvedContractContext, ResolvedEIP712Context
from erc7730.model.resolved.descriptor import ResolvedERC7730Descriptor
from erc7730.model.resolved.display import ResolvedDisplay, ResolvedFormat


@final
class ClassifyTransactionTypeLinter(ERC7730Linter):
    """- given schema/ABI, classify the transaction type
    - if class found, check descriptor display fields against predefined ruleset
    - possible class (swap, staking withdraw, staking deposit)
    """

    @override
    def lint(self, descriptor: ResolvedERC7730Descriptor, out: OutputAdder) -> None:
        if descriptor.context is None:
            return None
        if (tx_class := self._determine_tx_class(descriptor)) is None:
            # could not determine transaction type
            return None
        if (display := descriptor.display) is None:
            return None
        DisplayFormatChecker(tx_class, display).check(out)

    @classmethod
    def _determine_tx_class(cls, descriptor: ResolvedERC7730Descriptor) -> TxClass | None:
        if isinstance(descriptor.context, ResolvedEIP712Context):
            classifier = EIP712Classifier()
            if descriptor.context.eip712.schemas is not None:
                first_schema = descriptor.context.eip712.schemas[0]
                if isinstance(first_schema, EIP712Schema):
                    return classifier.classify(first_schema)
                # url should have been resolved earlier
        elif isinstance(descriptor.context, ResolvedContractContext):
            abi_classifier = ABIClassifier()
            if descriptor.context.contract.abi is not None:
                return abi_classifier.classify(descriptor.context.contract.abi)
        return None


class DisplayFormatChecker:
    """Given a transaction class and a display formats, check if all the required fields of a given
    transaction class are being displayed.
    If a field is missing emit an error.
    """

    def __init__(self, tx_class: TxClass, display: ResolvedDisplay):
        self.tx_class = tx_class
        self.display = display

    def check(self, out: OutputAdder) -> None:
        match self.tx_class:
            case TxClass.PERMIT:
                formats = self.display.formats
                fields = self._get_all_displayed_fields(formats)
                if not self._fields_contain("spender", fields):
                    out.warning(
                        title="Expected Display field missing",
                        message="Contract detected as Permit but no spender field displayed",
                    )
                if not self._fields_contain("amount", fields):
                    out.warning(
                        title="Expected Display field missing",
                        message="Contract detected as Permit but no amount field displayed",
                    )
                if (
                    not self._fields_contain("valid until", fields)
                    and not self._fields_contain("expiry", fields)
                    and not self._fields_contain("expiration", fields)
                    and not self._fields_contain("deadline", fields)
                ):
                    out.warning(
                        title="Expected Display field missing",
                        message="Contract detected as Permit but no expiration field displayed",
                    )
            case _:
                pass

    @classmethod
    def _get_all_displayed_fields(cls, formats: dict[str, ResolvedFormat]) -> set[str]:
        fields: set[str] = set()
        for format in formats.values():
            if format.fields is not None:
                for field in format.fields:
                    fields.add(str(field))
        return fields

    @classmethod
    def _fields_contain(cls, word: str, fields: set[str]) -> bool:
        """Check if the provided keyword is contained in one of the fields (case insensitive)"""
        return any(word.lower() in field.lower() for field in fields)



"""
Tooling for the ERC-7730 standard.
"""

# Base directory for descriptors in https://github.com/LedgerHQ/clear-signing-erc7730-registry.
ERC_7730_REGISTRY_DIRECTORY = "registry"

# File name prefix for EIP-712 descriptors in https://github.com/LedgerHQ/clear-signing-erc7730-registry.
ERC_7730_REGISTRY_EIP712_PREFIX = "eip712-"

# File name prefix for calldata descriptors in https://github.com/LedgerHQ/clear-signing-erc7730-registry.
ERC_7730_REGISTRY_CALLDATA_PREFIX = "calldata-"



from enum import Enum
from typing import Annotated, Any

from pydantic import Field, RootModel

from erc7730.model.base import Model
from erc7730.model.types import Id

# ruff: noqa: N815 - camel case field names are tolerated to match schema


class FieldFormat(str, Enum):
    """
    The format of the field, that will be used to format the field value in a human readable way.
    """

    RAW = "raw"
    """The field should be displayed as the natural representation of the underlying structured data type."""

    ADDRESS_NAME = "addressName"
    """The field should be displayed as a trusted name, or as a raw address if no names are found in trusted sources.
    List of trusted sources can be optionally specified in parameters."""

    CALL_DATA = "calldata"
    """The field is itself a calldata embedded in main call. Another ERC 7730 should be used to parse this field. If not
    available or not supported, the wallet MAY display a hash of the embedded calldata instead."""

    AMOUNT = "amount"
    """The field should be displayed as an amount in underlying currency, converted using the best magnitude / ticker
    available."""

    TOKEN_AMOUNT = "tokenAmount"  # nosec B105 - bandit false positive
    """The field should be displayed as an amount, preceded by the ticker. The magnitude and ticker should be derived
    from the tokenPath parameter corresponding metadata."""

    NFT_NAME = "nftName"
    """The field should be displayed as a single NFT names, or as a raw token Id if a specific name is not found.
    Collection is specified by the collectionPath parameter."""

    DATE = "date"
    """The field should be displayed as a date. Suggested RFC3339 representation. Parameter specifies the encoding of
    the date."""

    DURATION = "duration"
    """The field should be displayed as a duration in HH:MM:ss form. Value is interpreted as a number of seconds."""

    UNIT = "unit"
    """The field should be displayed as a percentage. Magnitude of the percentage encoding is specified as a parameter.
    Example: a value of 3000 with magnitude 4 is displayed as 0.3%."""

    ENUM = "enum"
    """The field should be displayed as a human readable string by converting the value using the enum referenced in
    parameters."""


class DateEncoding(str, Enum):
    """
    The encoding for a date.
    """

    BLOCKHEIGHT = "blockheight"
    """The date is encoded as a block height."""

    TIMESTAMP = "timestamp"
    """The date is encoded as a timestamp."""


class AddressNameType(str, Enum):
    """
    The type of address to display. Restrict allowable sources of names and MAY lead to additional checks from wallets.
    """

    WALLET = "wallet"
    """Address is an account controlled by the wallet."""

    EOA = "eoa"
    """Address is an Externally Owned Account."""

    CONTRACT = "contract"
    """Address is a well known smartcontract."""

    TOKEN = "token"  # nosec B105 - bandit false positive
    """Address is a well known ERC-20 token."""

    COLLECTION = "collection"
    """Address is a well known NFT collection."""


class Screen(RootModel[dict[str, Any]]):
    """
    Screens section is used to group multiple fields to display into screens. Each key is a wallet type name. The
    format of the screens is wallet type dependent, as well as what can be done (reordering fields, max number of
    screens, etc...). See each wallet manufacturer documentation for more information.
    """


SimpleIntent = Annotated[
    str,
    Field(
        title="Simple Intent",
        description="A description of the intent of the structured data signing, that will be displayed to the user.",
        min_length=1,
        max_length=256,  # TODO: arbitrary value, to be refined
    ),
]

ComplexIntent = Annotated[
    dict[str, str],
    Field(
        title="Complex Intent",
        description="A description of the intent of the structured data signing, that will be displayed to the user.",
        min_length=1,
        max_length=32,  # TODO: arbitrary value, to be refined
    ),
]


class FormatBase(Model):
    """
    A structured data format specification, containing formatting information of fields in a single type of message.
    """

    id: Id | None = Field(
        alias="$id",
        default=None,
        title="Id",
        description="An internal identifier that can be used either for clarity specifying what the element is or as a "
        "reference in device specific sections.",
    )

    intent: SimpleIntent | ComplexIntent | None = Field(
        default=None,
        title="Intent Message",
        description="A description of the intent of the structured data signing, that will be displayed to the user.",
    )

    screens: dict[str, list[Screen]] | None = Field(
        default=None,
        title="Screens grouping information",
        description="Screens section is used to group multiple fields to display into screens. Each key is a wallet "
        "type name. The format of the screens is wallet type dependent, as well as what can be done (reordering "
        "fields, max number of screens, etc...). See each wallet manufacturer documentation for more "
        "information.",
    )



from typing import assert_never, final, override

from eip712.model.resolved.descriptor import ResolvedEIP712DAppDescriptor
from eip712.model.resolved.message import ResolvedEIP712MapperField
from eip712.model.schema import EIP712SchemaField, EIP712Type
from eip712.model.types import EIP712Format
from eip712.utils import MissingRootTypeError, MultipleRootTypesError, get_primary_type
from pydantic_string_url import HttpUrl

from erc7730.common.output import ExceptionsToOutput, OutputAdder
from erc7730.convert import ERC7730Converter
from erc7730.model.context import EIP712Schema
from erc7730.model.display import (
    DateEncoding,
    FieldFormat,
)
from erc7730.model.input.context import InputDeployment, InputDomain, InputEIP712, InputEIP712Context
from erc7730.model.input.descriptor import InputERC7730Descriptor
from erc7730.model.input.display import (
    InputDateParameters,
    InputDisplay,
    InputFieldDescription,
    InputFormat,
    InputNestedFields,
    InputReference,
    InputTokenAmountParameters,
)
from erc7730.model.input.metadata import InputMetadata
from erc7730.model.paths import ContainerField, ContainerPath


@final
class EIP712toERC7730Converter(ERC7730Converter[ResolvedEIP712DAppDescriptor, InputERC7730Descriptor]):
    """
    Converts Ledger legacy EIP-712 descriptor to ERC-7730 descriptor.

    Generates 1 output ERC-7730 descriptor per contract, as ERC-7730 descriptors only represent 1 contract.
    """

    @override
    def convert(
        self, descriptor: ResolvedEIP712DAppDescriptor, out: OutputAdder
    ) -> dict[str, InputERC7730Descriptor] | None:
        with ExceptionsToOutput(out):
            descriptors: dict[str, InputERC7730Descriptor] = {}

            for contract in descriptor.contracts:
                formats: dict[str, InputFormat] = {}
                schemas: list[EIP712Schema | HttpUrl] = []

                for message in contract.messages:
                    if (primary_type := self._get_primary_type(message.schema_, out)) is None:
                        return None

                    schemas.append(EIP712Schema(primaryType=primary_type, types=message.schema_))

                    formats[primary_type] = InputFormat(
                        intent=message.mapper.label,
                        fields=[self._convert_field(field) for field in message.mapper.fields],
                        required=None,
                        screens=None,
                    )

                descriptors[contract.address] = InputERC7730Descriptor(
                    context=InputEIP712Context(
                        eip712=InputEIP712(
                            domain=InputDomain(
                                name=descriptor.name,
                                version=None,
                                chainId=descriptor.chainId,
                                verifyingContract=contract.address,
                            ),
                            schemas=schemas,
                            deployments=[InputDeployment(chainId=descriptor.chainId, address=contract.address)],
                        )
                    ),
                    metadata=InputMetadata(
                        owner=contract.contractName,
                        info=None,
                        token=None,
                        constants=None,
                        enums=None,
                    ),
                    display=InputDisplay(
                        definitions=None,
                        formats=formats,
                    ),
                )

        return descriptors

    @classmethod
    def _convert_field(
        cls, field: ResolvedEIP712MapperField
    ) -> InputFieldDescription | InputReference | InputNestedFields:
        # FIXME must generate nested fields for arrays
        match field.format:
            case EIP712Format.RAW | None:
                return InputFieldDescription(path=field.path, label=field.label, format=FieldFormat.RAW, params=None)
            case EIP712Format.AMOUNT if field.assetPath is not None:
                return InputFieldDescription(
                    path=field.path,
                    label=field.label,
                    format=FieldFormat.TOKEN_AMOUNT,
                    params=InputTokenAmountParameters(tokenPath=field.assetPath),
                )
            case EIP712Format.AMOUNT:
                return InputFieldDescription(
                    path=field.path,
                    label=field.label,
                    format=FieldFormat.TOKEN_AMOUNT,
                    params=InputTokenAmountParameters(tokenPath=ContainerPath(field=ContainerField.TO)),
                )
            case EIP712Format.DATETIME:
                return InputFieldDescription(
                    path=field.path,
                    label=field.label,
                    format=FieldFormat.DATE,
                    params=InputDateParameters(encoding=DateEncoding.TIMESTAMP),
                )
            case _:
                assert_never(field.format)

    @classmethod
    def _get_primary_type(
        cls, schema: dict[EIP712Type, list[EIP712SchemaField]], out: OutputAdder
    ) -> EIP712Type | None:
        try:
            return get_primary_type(schema)
        except MissingRootTypeError:
            return out.error(
                title="Invalid EIP-712 schema",
                message="Primary type could not be determined on EIP-712 schema, as all types are referenced by"
                "other types. Please make sure your schema has a root type.",
            )
        except MultipleRootTypesError:
            return out.error(
                title="Invalid EIP-712 schema",
                message="Primary type could not be determined on EIP-712 schema, as several types are not"
                "referenced by any other type. Please make sure your schema has a single root type.",
            )



from abc import ABC, abstractmethod

from erc7730.common.output import OutputAdder
from erc7730.model.resolved.descriptor import ResolvedERC7730Descriptor


class ERC7730Linter(ABC):
    """
    Linter for ERC-7730 descriptors, inspects a (structurally valid) descriptor and emits notes, warnings, or
    errors.

    A linter may emit false positives or false negatives. It is up to the user to interpret the output.
    """

    @abstractmethod
    def lint(self, descriptor: ResolvedERC7730Descriptor, out: OutputAdder) -> None:
        raise NotImplementedError()




import argparse
import json
import requests
import time
from pathlib import Path

parser = argparse.ArgumentParser()
parser.add_argument('-w', '--web3-handler', action="store_true", help="use web3-handler to get the abis files", default=True)
parser.add_argument('-l', '--local-web3-handler', action="store_true", help="use web3-handler to get the abis files", default=False)


COMPOUND_REPO_PATH = "external_ressources/compound-config/"
DAL_HOME_PATH = "../../"

ABI_ENDPOINT = 'https://api.etherscan.io/api?module=contract&action=getabi&address='
def get_etherscan_abi(contract_address):
    while True:
        time.sleep(0.3)

        # Get the ABI
        response = requests.get('%s%s'%(ABI_ENDPOINT, contract_address))
        response_json = response.json()

        #Failed! try again
        if (response_json['status'] == '0'):
            if (response_json['result'] == "Max rate limit reached, please use API Key for higher rate limit"):
                print("Failed to fetch " + str(contract_address) + " ABI. Retrying...")
                continue
            else:
                return None # Echec total!

        #Success! return abi
        abi_json = json.loads(response_json['result'])
        return abi_json

# Web3handler return contract abi + the implementation ABI behind the proxy
PROD_WEB3_HANDLER_ABI_FSTRING = 'https://web3-handler-api-prd2.aws.prd.ldg-tech.com/%s/%s/abi'
LOCAL_WEB3_HANDLER_ABI_FSTRING = 'http://localhost:9003/%s/%s/abi'
def get_web3handler_abi(blockchain, contract_address, use_local_web3_handler):
    time.sleep(0.3)
    print(contract_address)
    endpoint_fstring = LOCAL_WEB3_HANDLER_ABI_FSTRING if use_local_web3_handler else PROD_WEB3_HANDLER_ABI_FSTRING
    response = requests.get(endpoint_fstring%(blockchain, contract_address)).json()
    if (type(response) is dict and "detail" in response.keys()):
        if ("detail" not in response.keys()):
            raise Exception("Unexpected web3_handler response: " + str(response))

        if response['detail'].startswith("ABI was not found for address"):
            print("WEB3 HANDLER:" + str(response['detail']))
            return None
        elif response['detail'].startswith("Invalid address format:"):
            raise Exception("Unexpected web3_handler response: " + str(response))
        else:
            raise Exception("Unexpected web3_handler response: " + str(response))

    if (type(response) is not list):
        raise Exception("Unexpected web3_handler response: " + str(response))

    # return ABI
    return response



class ConfigFile:
    def __init__(self, cal_network_name, compound_contract_file, compound_abi_file):
        self.network_name = cal_network_name
        self.cal_abi_folder_path = Path(DAL_HOME_PATH) / cal_network_name / "compound" / "abis"
        if (self.cal_abi_folder_path.is_dir() == False):
            print("Creating folder " + str(self.cal_abi_folder_path))
            self.cal_abi_folder_path.mkdir(mode=0o775, parents=True, exist_ok=True)
        else:
            print("Found folder " + str(self.cal_abi_folder_path))

        self.contract_list_file = Path(DAL_HOME_PATH) / COMPOUND_REPO_PATH / 'networks' / Path(compound_contract_file)
        assert(self.contract_list_file.exists())

        self.abi_list_file = Path(DAL_HOME_PATH) / COMPOUND_REPO_PATH / 'networks' / Path(compound_abi_file)
        assert(self.abi_list_file.exists())


    def fill_abi_files(self, use_web3_handler, use_local_web3_handler):
        with self.contract_list_file.open() as contracts_fd, \
             self.abi_list_file.open() as abis_fd:
            contracts_json = json.load(contracts_fd)
            abis_json = json.load(abis_fd)

            failed_contracts = []

            # Get ABI for each contracts
            triplets = []
            for contract_name, contract_address in contracts_json["Contracts"].items():
                abi = None
                if (use_web3_handler):
                    print("Fetching : " + contract_name + "(" + contract_address + ") from web3_handler: ", end='')
                    abi = get_web3handler_abi(self.network_name, contract_address, use_local_web3_handler)

                # if ABI is not found on web3_handler search official abi then on etherscan
                if abi is None:
                    if contract_name not in abis_json.keys():
                        print("/!\\ Warning: " + contract_name + "(" + contract_address + ") ABI not found in listed compound repository. Fetching Etherscan ABI")
                        abi = get_etherscan_abi(contract_address)
                    else:
                        print("ABI found in compound repository for " + contract_name + "(" + contract_address + ")")
                        abi = abis_json[contract_name]

                if (abi is not None):
                    triplets.append((contract_name, contract_address, abi))
                else:
                    print("Cannot find abi for contract " + contract_name + "(" + contract_address + ")" )
                    failed_contracts.append((contract_name, contract_address, abi))


            # Export ABI files
            for contract_name, contract_address, abi in triplets:
                cal_abi_file_path = (self.cal_abi_folder_path / contract_address.lower()).with_suffix(".abi.json")
                print("Writing " + str(cal_abi_file_path))
                with cal_abi_file_path.open(mode="w") as cal_abi_file_path_fd:
                    json.dump(abi, cal_abi_file_path_fd, indent=4, sort_keys=True)

            # Reporting
            total_count = len(triplets) + len(failed_contracts)
            print("===================================================================================================================")
            print("Network: " + str(self.network_name))
            print("Result {0}/{1} succeeded {2}/{1} failed.".format(len(triplets), total_count, len(failed_contracts)))
            for contract_name, contract_address, abi in failed_contracts:
                print ("Failed to get ABI for " + contract_name + "(" + contract_address + ")")
            print("===================================================================================================================")


networks = [
    ConfigFile(cal_network_name='ethereum', compound_contract_file='mainnet.json', compound_abi_file='mainnet-abi.json'),
    ConfigFile(cal_network_name='ethereum_goerli', compound_contract_file='goerli.json', compound_abi_file='goerli-abi.json'),
    # ConfigFile(cal_network_name='ethereum_ropsten', compound_contract_file='ropsten.json', compound_abi_file='ropsten-abi.json'),
]

def main():
    args = parser.parse_args()

    for network in networks:
        network.fill_abi_files(args.web3_handler or args.local_web3_handler, args.local_web3_handler)

if __name__ == "__main__":
    main()



from typing import assert_never, final, override

from eip712.model.input.contract import InputEIP712Contract
from eip712.model.input.descriptor import InputEIP712DAppDescriptor
from eip712.model.input.message import InputEIP712Mapper, InputEIP712MapperField, InputEIP712Message
from eip712.model.schema import EIP712SchemaField
from eip712.model.types import EIP712Format

from erc7730.common.ledger import ledger_network_id
from erc7730.common.output import ExceptionsToOutput, OutputAdder
from erc7730.convert import ERC7730Converter
from erc7730.model.context import EIP712Schema
from erc7730.model.display import FieldFormat
from erc7730.model.paths import Array, ContainerField, ContainerPath, DataPath
from erc7730.model.paths.path_ops import data_path_concat, to_relative
from erc7730.model.resolved.context import ResolvedDeployment, ResolvedEIP712Context
from erc7730.model.resolved.descriptor import ResolvedERC7730Descriptor
from erc7730.model.resolved.display import (
    ResolvedField,
    ResolvedFieldDescription,
    ResolvedNestedFields,
    ResolvedTokenAmountParameters,
    ResolvedValueConstant,
    ResolvedValuePath,
)


@final
class ERC7730toEIP712Converter(ERC7730Converter[ResolvedERC7730Descriptor, InputEIP712DAppDescriptor]):
    """
    Converts ERC-7730 descriptor to Ledger legacy EIP-712 descriptor.

    Generates 1 output InputEIP712DAppDescriptor per chain id, as EIP-712 descriptors are chain-specific.
    """

    @override
    def convert(
        self, descriptor: ResolvedERC7730Descriptor, out: OutputAdder
    ) -> dict[str, InputEIP712DAppDescriptor] | None:
        with ExceptionsToOutput(out):
            context = descriptor.context
            if not isinstance(context, ResolvedEIP712Context):
                return out.error("context is not EIP712")

            if (domain := context.eip712.domain) is None or (dapp_name := domain.name) is None:
                return out.error("EIP712 domain is not defined")

            if (contract_name := descriptor.metadata.owner) is None:
                return out.error("metadata.owner is not defined")

            messages: list[InputEIP712Message] = []
            for primary_type, format in descriptor.display.formats.items():
                schema = self._get_schema(primary_type, context.eip712.schemas, out)

                if schema is None:
                    return out.error(f"EIP-712 schema for {primary_type} is missing")

                label = format.intent if isinstance(format.intent, str) else primary_type

                output_fields = []
                for input_field in format.fields:
                    if (out_field := self.convert_field(input_field, None, out)) is None:
                        return None
                    output_fields.extend(out_field)

                messages.append(
                    InputEIP712Message(schema=schema, mapper=InputEIP712Mapper(label=label, fields=output_fields))
                )

            descriptors: dict[str, InputEIP712DAppDescriptor] = {}
            for deployment in context.eip712.deployments:
                output_descriptor = self._build_network_descriptor(deployment, dapp_name, contract_name, messages, out)
                if output_descriptor is not None:
                    descriptors[str(deployment.chainId)] = output_descriptor

        return descriptors

    @classmethod
    def _build_network_descriptor(
        cls,
        deployment: ResolvedDeployment,
        dapp_name: str,
        contract_name: str,
        messages: list[InputEIP712Message],
        out: OutputAdder,
    ) -> InputEIP712DAppDescriptor | None:
        if (network := ledger_network_id(deployment.chainId)) is None:
            return out.error(f"network id {deployment.chainId} not supported")

        return InputEIP712DAppDescriptor(
            blockchainName=network,
            chainId=deployment.chainId,
            name=dapp_name,
            contracts=[
                InputEIP712Contract(address=deployment.address.lower(), contractName=contract_name, messages=messages)
            ],
        )

    @classmethod
    def _get_schema(
        cls, primary_type: str, schemas: list[EIP712Schema], out: OutputAdder
    ) -> dict[str, list[EIP712SchemaField]] | None:
        for schema in schemas:
            if schema.primaryType == primary_type:
                return schema.types
        return out.error(f"schema for type {primary_type} not found")

    @classmethod
    def convert_field(
        cls, field: ResolvedField, prefix: DataPath | None, out: OutputAdder
    ) -> list[InputEIP712MapperField] | None:
        match field:
            case ResolvedFieldDescription():
                if (output_field := cls.convert_field_description(field, prefix, out)) is None:
                    return None
                return [output_field]
            case ResolvedNestedFields():
                output_fields = []
                for in_field in field.fields:
                    if (output_field := cls.convert_field(in_field, prefix, out)) is None:
                        return None
                    output_fields.extend(output_field)
                return output_fields
            case _:
                assert_never(field)

    @classmethod
    def convert_field_description(
        cls,
        field: ResolvedFieldDescription,
        prefix: DataPath | None,
        out: OutputAdder,
    ) -> InputEIP712MapperField | None:
        field_path: DataPath
        asset_path: DataPath | None = None
        field_format: EIP712Format | None = None
        in_array: bool = False

        match field.value:
            case ResolvedValueConstant():
                return out.error("Constant values are not supported")

            case ResolvedValuePath(path=path):
                match path:
                    case DataPath() as field_path:
                        field_path = data_path_concat(prefix, field_path)

                        for element in field_path.elements:
                            match element:
                                case Array():
                                    in_array = True
                                    break
                                case _:
                                    pass

                    case ContainerPath() as container_path:
                        return out.error(f"Path {container_path} is not supported")
                    case _:
                        assert_never(field.value)
            case _:
                assert_never(field.value)

        match field.format:
            case None:
                field_format = None
            case FieldFormat.ADDRESS_NAME:
                field_format = EIP712Format.RAW
            case FieldFormat.RAW:
                field_format = EIP712Format.RAW
            case FieldFormat.ENUM:
                field_format = EIP712Format.RAW
            case FieldFormat.UNIT:
                field_format = EIP712Format.RAW
            case FieldFormat.DURATION:
                field_format = EIP712Format.RAW
            case FieldFormat.NFT_NAME:
                field_format = EIP712Format.RAW
            case FieldFormat.CALL_DATA:
                field_format = EIP712Format.RAW
            case FieldFormat.DATE:
                field_format = EIP712Format.DATETIME
            case FieldFormat.AMOUNT:
                field_format = EIP712Format.AMOUNT
            case FieldFormat.TOKEN_AMOUNT:
                if in_array:
                    # EIP-712 does not support token references in arrays, fallback to raw format
                    field_format = EIP712Format.RAW
                else:
                    field_format = EIP712Format.AMOUNT
                    if field.params is not None and isinstance(field.params, ResolvedTokenAmountParameters):
                        match field.params.token:
                            case None:
                                return out.error("Token path or reference must be set")

                            case ResolvedValueConstant():
                                return out.error("Constant values are not supported")

                            case ResolvedValuePath(path=path):
                                match path:
                                    case None:
                                        pass
                                    case DataPath() as token_path:
                                        asset_path = data_path_concat(prefix, token_path)
                                    case ContainerPath() as container_path if container_path.field == ContainerField.TO:
                                        # In EIP-712 protocol, format=token with no token path
                                        #  => refers to verifyingContract
                                        asset_path = None
                                    case ContainerPath() as container_path:
                                        return out.error(f"Path {container_path} is not supported")
                                    case _:
                                        assert_never(path)
                            case _:
                                assert_never(field.value)
            case _:
                assert_never(field.format)

        return InputEIP712MapperField(
            path=str(to_relative(field_path)),
            label=field.label,
            assetPath=None if asset_path is None else str(to_relative(asset_path)),
            format=field_format,
        )



import pytest
from eip712.model.input.descriptor import InputEIP712DAppDescriptor

from erc7730.model.input.descriptor import InputERC7730Descriptor
from tests.assertions import assert_model_json_schema
from tests.files import ERC7730_SCHEMA, LEGACY_REGISTRY
from tests.io import load_json_file


def assert_valid_erc_7730(descriptor: InputERC7730Descriptor) -> None:
    """Assert descriptor serializes to a JSON that passes JSON schema validation."""
    assert_model_json_schema(descriptor, ERC7730_SCHEMA)


def assert_valid_legacy_eip_712(descriptor: InputEIP712DAppDescriptor) -> None:
    """Assert descriptor serializes to a JSON that passes JSON schema validation."""
    schema_path = LEGACY_REGISTRY / descriptor.blockchainName / "eip712.schema.json"
    if not schema_path.is_file():
        pytest.skip(f"Legacy registry has no EIP-712 schema for network {descriptor.blockchainName}")
    schema = load_json_file(schema_path)
    assert_model_json_schema(descriptor, schema)



import contextlib
import operator
from collections.abc import Callable
from functools import reduce
from typing import Any


def get_by_path(root: dict[str, Any], *paths: str) -> Any:
    """Access a nested object in root by item sequence."""
    return reduce(operator.getitem, list(paths), root)


def is_in_path(root: dict[str, Any], *paths: str) -> bool:
    """Check if a key-value in a nested object in root by item sequence exists."""
    try:
        get_by_path(root, *paths)
        return True
    except KeyError:
        return False


def del_by_path(root: dict[str, Any], *paths: str) -> None:
    """Delete a key-value in a nested object in root by item sequence."""
    path_list = list(paths)
    with contextlib.suppress(KeyError):
        del get_by_path(root, *path_list[:-1])[path_list[-1]]


def map_by_path(root: dict[str, Any], func: Callable[[Any], Any], *paths: str) -> None:
    """Map a function to a key-value in a nested object in root by item sequence."""
    path_list = list(paths)
    with contextlib.suppress(KeyError):
        get_by_path(root, *path_list[:-1])[path_list[-1]] = func(get_by_path(root, *paths))



import json
from pathlib import Path
from typing import Any

import jsonschema
from prettydiff import print_diff

from erc7730.common.pydantic import _BaseModel, model_to_json_str


def assert_dict_equals(expected: dict[str, Any], actual: dict[str, Any]) -> None:
    """Assert dictionaries are equal, pretty printing differences to console."""
    if expected != actual:
        print_diff(expected, actual)
        raise AssertionError("Dictionaries are not equal")


def assert_json_str_equals(expected: str, actual: str) -> None:
    """Assert deserialized JSON strings are equal."""
    assert_dict_equals(json.loads(expected), json.loads(actual))


def assert_json_file_equals(expected: Path, actual: Path) -> None:
    """Assert deserialized JSON files are equal."""
    with open(expected) as exp, open(actual) as act:
        assert_dict_equals(json.load(exp), json.load(act))


def assert_model_json_equals(expected: _BaseModel, actual: _BaseModel) -> None:
    """Assert models serialize to same JSON, pretty printing differences to console."""
    assert_json_str_equals(model_to_json_str(expected), model_to_json_str(actual))


def assert_model_json_schema(model: _BaseModel, schema: Any) -> None:
    """Assert model is valid against JSON schema."""
    try:
        jsonschema.validate(instance=json.loads(model_to_json_str(model)), schema=schema)
    except jsonschema.exceptions.ValidationError as e:
        raise AssertionError("Document does not respect JSON schema") from e



from pathlib import Path

from rich import print

from erc7730.common.output import ConsoleOutputAdder, RaisingOutputAdder
from erc7730.common.pydantic import model_to_json_file
from erc7730.convert import ERC7730Converter, InputType, OutputType


def convert_to_file_and_print_errors(
    input_descriptor: InputType, output_file: Path, converter: ERC7730Converter[InputType, OutputType]
) -> bool:
    """
    Convert an input descriptor to an output file using a converter, and print any errors encountered.

    :param input_descriptor: loaded, valid input descriptor
    :param output_file: output file path (overwritten if already exists)
    :param converter: converter to use
    :return: True if output file was written (if no errors, or only non-fatal errors encountered)
    """
    if (output_descriptor := convert_and_print_errors(input_descriptor, converter)) is not None:
        if isinstance(output_descriptor, dict):
            for identifier, descriptor in output_descriptor.items():
                descriptor_file = output_file.with_suffix(f".{identifier}{output_file.suffix}")
                model_to_json_file(descriptor_file, descriptor)
                print(f"[green]generated {descriptor_file} ✅[/green]")
        else:
            model_to_json_file(output_file, output_descriptor)
            print(f"[green]generated {output_file} ✅[/green]")
        return True

    print("[red]conversion failed ❌[/red]")
    return False


def convert_and_print_errors(
    input_descriptor: InputType, converter: ERC7730Converter[InputType, OutputType]
) -> OutputType | dict[str, OutputType] | None:
    """
    Convert an input descriptor using a converter, print any errors encountered, and return the result model.

    :param input_descriptor: loaded, valid input descriptor
    :param converter: converter to use
    :return: output descriptor (if no errors, or only non-fatal errors encountered), None otherwise
    """
    return _normalize_result(converter.convert(input_descriptor, ConsoleOutputAdder()))


def convert_and_raise_errors(
    input_descriptor: InputType, converter: ERC7730Converter[InputType, OutputType]
) -> OutputType | dict[str, OutputType] | None:
    """
    Convert an input descriptor using a converter, raising any errors encountered, and return the result model.

    :param input_descriptor: loaded, valid input descriptor
    :param converter: converter to use
    :return: output descriptor (if no errors, or only non-fatal errors encountered), None otherwise
    """
    return _normalize_result(converter.convert(input_descriptor, RaisingOutputAdder()))


def _normalize_result(result: OutputType | dict[str, OutputType] | None) -> OutputType | dict[str, OutputType] | None:
    if isinstance(result, dict):
        match len(result):
            case 0:
                return None
            case 1:
                return next(iter(result.values()))
            case _:
                return result
    return result



from pydantic_string_url import HttpUrl

from erc7730.common import client
from erc7730.model.abi import ABI


def test_get_supported_chains() -> None:
    result = client.get_supported_chains()
    assert result is not None
    assert len(result) >= 50
    names = {chain.chainname for chain in result}
    assert "Ethereum Mainnet" in names
    assert "Sepolia Testnet" in names
    assert "Holesky Testnet" in names
    assert "BNB Smart Chain Mainnet" in names
    assert "BNB Smart Chain Testnet" in names
    assert "Polygon Mainnet" in names
    assert "Polygon Amoy Testnet" in names
    assert "Polygon zkEVM Mainnet" in names
    assert "Polygon zkEVM Cardona Testnet" in names
    assert "Base Mainnet" in names
    assert "Base Sepolia Testnet" in names
    assert "Arbitrum One Mainnet" in names
    assert "Arbitrum Nova Mainnet" in names
    assert "Arbitrum Sepolia Testnet" in names
    assert "Linea Mainnet" in names
    assert "Linea Sepolia Testnet" in names
    assert "Fantom Opera Mainnet" in names
    assert "Fantom Testnet" in names
    assert "Blast Mainnet" in names
    assert "Blast Sepolia Testnet" in names
    assert "OP Mainnet" in names
    assert "OP Sepolia Testnet" in names
    assert "Avalanche C-Chain" in names
    assert "Avalanche Fuji Testnet" in names
    assert "BitTorrent Chain Mainnet" in names
    assert "BitTorrent Chain Testnet" in names
    assert "Celo Mainnet" in names
    assert "Celo Alfajores Testnet" in names
    assert "Cronos Mainnet" in names
    assert "Fraxtal Mainnet" in names
    assert "Fraxtal Testnet" in names
    assert "Gnosis" in names
    assert "Kroma Mainnet" in names
    assert "Kroma Sepolia Testnet" in names
    assert "Mantle Mainnet" in names
    assert "Mantle Sepolia Testnet" in names
    assert "Moonbeam Mainnet" in names
    assert "Moonriver Mainnet" in names
    assert "Moonbase Alpha Testnet" in names
    assert "opBNB Mainnet" in names
    assert "opBNB Testnet" in names
    assert "Scroll Mainnet" in names
    assert "Scroll Sepolia Testnet" in names
    assert "Taiko Mainnet" in names
    assert "Taiko Hekla L2 Testnet" in names
    assert "WEMIX3.0 Mainnet" in names
    assert "WEMIX3.0 Testnet" in names
    assert "zkSync Mainnet" in names
    assert "zkSync Sepolia Testnet" in names
    assert "Xai Mainnet" in names
    assert "Xai Sepolia Testnet" in names


def test_get_contract_abis() -> None:
    result = client.get_contract_abis(chain_id=1, contract_address="0x06012c8cf97bead5deae237070f9587f8e7a266d")
    assert result is not None
    assert len(result) > 0


def test_get_from_github() -> None:
    result1 = client.get(
        url=HttpUrl(
            "https://github.com/LedgerHQ/ledger-asset-dapps/blob/main"
            "/ethereum/uniswap/abis/0x000000000022d473030f116ddee9f6b43ac78ba3.abi.json"
        ),
        model=list[ABI],
    )
    result2 = client.get(
        url=HttpUrl(
            "https://raw.githubusercontent.com/LedgerHQ/ledger-asset-dapps/refs/heads/main"
            "/ethereum/uniswap/abis/0x000000000022d473030f116ddee9f6b43ac78ba3.abi.json"
        ),
        model=list[ABI],
    )
    assert result1 is not None
    assert result2 is not None
    assert len(result1) > 0
    assert len(result2) > 0
    assert result1 == result2



import json
from pathlib import Path
from typing import Any


def load_json_file(path: Path) -> Any:
    """Load and deserialize a JSON file."""
    with open(path, "rb") as file:
        return json.load(file)



from typing import Annotated, ForwardRef, Literal

from eip712.model.schema import EIP712Type
from pydantic import Discriminator, Field, Tag

from erc7730.common.abi import ABIDataType
from erc7730.model.base import Model
from erc7730.model.display import (
    AddressNameType,
    DateEncoding,
    FieldFormat,
    FormatBase,
)
from erc7730.model.paths import ContainerPath, DataPath
from erc7730.model.resolved.path import ResolvedPath
from erc7730.model.types import Address, HexStr, Id, ScalarType, Selector
from erc7730.model.unions import field_discriminator, field_parameters_discriminator

# ruff: noqa: N815 - camel case field names are tolerated to match schema


class ResolvedValuePath(Model):
    """
    A path to the field in the structured data. The path is a JSON path expression that can be used to extract the
    field value from the structured data.
    """

    type: Literal["path"] = Field(
        default="path",
        title="Value Type",
        description="The value type identifier (discriminator for values discriminated union).",
    )

    path: ResolvedPath = Field(
        title="Path",
        description="A path to the field in the structured data. The path is a JSON path expression that can be used "
        "to extract the field value from the structured data.",
    )


class ResolvedValueConstant(Model):
    """
    A constant value.
    """

    type: Literal["constant"] = Field(
        default="constant",
        title="Value Type",
        description="The value type identifier (discriminator for values discriminated union).",
    )

    type_family: ABIDataType = Field(
        title="Type family",
        description="Type family of the value",
    )

    type_size: int | None = Field(
        default=None, title="Type size", description="Size of the type (in bytes)", ge=0, le=255
    )

    value: ScalarType = Field(
        title="Value",
        description="The constant value.",
    )

    raw: HexStr = Field(
        title="Raw Value",
        description="The constant value, serialized and hex encoded.",
    )


ResolvedValue = Annotated[
    ResolvedValuePath | ResolvedValueConstant,
    Discriminator("type"),
]


class ResolvedTokenAmountParameters(Model):
    """
    Token Amount Formatting Parameters.
    """

    token: ResolvedValue | None = Field(
        default=None,
        title="Token",
        description="The address of the token contract, either as path to a field in the structured data or a constant "
        "value. Used to associate correct ticker. If ticker is not found or value is not set, the wallet "
        'SHOULD display the raw value instead with an "Unknown token" warning.',
    )

    nativeCurrencyAddress: list[Address] | None = Field(
        default=None,
        title="Native Currency Address",
        description="An address or array of addresses, any of which are interpreted as an amount in native currency "
        "rather than a token.",
    )

    threshold: HexStr | None = Field(
        default=None,
        title="Unlimited Threshold",
        description="The threshold above which the amount should be displayed using the message parameter rather than "
        "the real amount (encoded as a byte array).",
    )

    message: str | None = Field(
        default=None,
        title="Unlimited Message",
        description="The message to display when the amount is above the threshold.",
    )


class ResolvedAddressNameParameters(Model):
    """
    Address Names Formatting Parameters.
    """

    types: list[AddressNameType] | None = Field(
        default=None,
        title="Address Type",
        description="An array of expected types of the address. If set, the wallet SHOULD check that the address "
        "matches one of the types provided.",
        min_length=1,
    )

    sources: list[str] | None = Field(
        default=None,
        title="Trusted Sources",
        description="An array of acceptable sources for names. If set, the wallet SHOULD restrict name lookup to "
        "relevant sources.",
        min_length=1,
    )


class ResolvedCallDataParameters(Model):
    """
    Embedded Calldata Formatting Parameters.
    """

    selector: str | None = Field(
        default=None,
        title="Called Selector",
        description="The selector being called, if not contained in the calldata. Hex string representation.",
    )

    callee: ResolvedValue = Field(
        title="Callee",
        description="The address of the contract being called by this embedded calldata, either as path to a field in "
        "the structured data or a constant value.",
    )


class ResolvedNftNameParameters(Model):
    """
    NFT Names Formatting Parameters.
    """

    collection: ResolvedValue = Field(
        title="Collection",
        description="The address of the collection contract, either as path to a field in the structured data or a "
        "constant value.",
    )


class ResolvedDateParameters(Model):
    """
    Date Formatting Parameters
    """

    encoding: DateEncoding = Field(title="Date Encoding", description="The encoding of the date.")


class ResolvedUnitParameters(Model):
    """
    Unit Formatting Parameters.
    """

    base: str = Field(
        title="Unit base symbol",
        description="The base symbol of the unit, displayed after the converted value. It can be an SI unit symbol or "
        "acceptable dimensionless symbols like % or bps.",
    )

    decimals: int | None = Field(
        default=None,
        title="Decimals",
        description="The number of decimals of the value, used to convert to a float.",
        ge=0,
        le=255,
    )

    prefix: bool | None = Field(
        default=None,
        title="Prefix",
        description="Whether the value should be converted to a prefixed unit, like k, M, G, etc.",
    )


class ResolvedEnumParameters(Model):
    """
    Enum Formatting Parameters.
    """

    enumId: Id = Field(
        title="Enum Identifier",
        description="The identifier of the enum in the $.context.metadata section.",
    )


ResolvedFieldParameters = Annotated[
    Annotated[ResolvedAddressNameParameters, Tag("address_name")]
    | Annotated[ResolvedCallDataParameters, Tag("call_data")]
    | Annotated[ResolvedTokenAmountParameters, Tag("token_amount")]
    | Annotated[ResolvedNftNameParameters, Tag("nft_name")]
    | Annotated[ResolvedDateParameters, Tag("date")]
    | Annotated[ResolvedUnitParameters, Tag("unit")]
    | Annotated[ResolvedEnumParameters, Tag("enum")],
    Discriminator(field_parameters_discriminator),
]


class ResolvedFieldBase(Model):
    """
    A field formatter, containing formatting information of a single field in a message.
    """

    value: ResolvedValue = Field(
        title="Value",
        description="A reference to the value to display, either as path to a field in the structured data or a "
        "constant value.",
    )


class ResolvedFieldDescription(ResolvedFieldBase):
    """
    A field formatter, containing formatting information of a single field in a message.
    """

    id: Id | None = Field(
        alias="$id",
        default=None,
        title="Id",
        description="An internal identifier that can be used either for clarity specifying what the element is or as a"
        "reference in device specific sections.",
    )

    label: str = Field(
        title="Field Label",
        description="The label of the field, that will be displayed to the user in front of the formatted field value.",
    )

    format: FieldFormat | None = Field(
        title="Field Format",
        description="The format of the field, that will be used to format the field value in a human readable way.",
    )

    params: ResolvedFieldParameters | None = Field(
        default=None,
        title="Format Parameters",
        description="Format specific parameters that are used to format the field value in a human readable way.",
    )


class ResolvedNestedFields(ResolvedFieldBase):
    """
    A single set of field formats, allowing recursivity in the schema.

    Used to group whole definitions for structures for instance. This allows nesting definitions of formats, but note
    that support for deep nesting will be device dependent.
    """

    fields: list[ForwardRef("ResolvedField")] = Field(  # type: ignore
        title="Fields", description="Nested fields formats."
    )


ResolvedField = Annotated[
    Annotated[ResolvedFieldDescription, Tag("field_description")]
    | Annotated[ResolvedNestedFields, Tag("nested_fields")],
    Discriminator(field_discriminator),
]


class ResolvedFormat(FormatBase):
    """
    A structured data format specification, containing formatting information of fields in a single type of message.
    """

    fields: list[ResolvedField] = Field(
        title="Field Formats set", description="An array containing the ordered definitions of fields formats."
    )

    required: list[DataPath | ContainerPath] | None = Field(
        default=None,
        title="Required fields",
        description="A list of fields that are required to be displayed to the user. A field that has a formatter and "
        "is not in this list is optional. A field that does not have a formatter should be silent, ie not "
        "shown.",
    )

    excluded: list[DataPath] | None = Field(
        default=None,
        title="Excluded fields",
        description="Intentionally excluded fields, as an array of *paths* referring to specific fields. A field that "
        "has no formatter and is not declared in this list MAY be considered as an error by the wallet when "
        "interpreting the descriptor. The excluded paths should interpreted as prefixes, meaning that all fields under "
        "excluded path should be ignored",
    )


class ResolvedDisplay(Model):
    """
    Display Formatting Info Section.
    """

    formats: dict[EIP712Type | Selector, ResolvedFormat] = Field(
        title="List of field formats",
        description="The list includes formatting info for each field of a structure. This list is indexed by a key"
        "identifying uniquely the message's type in the abi. For smartcontracts, it is the selector of the"
        "function or its signature; and for EIP712 messages it is the primaryType of the message.",
    )



from typing import final, override

from erc7730.common.output import OutputAdder
from erc7730.lint import ERC7730Linter
from erc7730.model.resolved.descriptor import ResolvedERC7730Descriptor


@final
class MultiLinter(ERC7730Linter):
    """A linter that runs multiple linters in sequence."""

    def __init__(self, linters: list[ERC7730Linter]):
        self.lints = linters

    @override
    def lint(self, descriptor: ResolvedERC7730Descriptor, out: OutputAdder) -> None:
        for linter in self.lints:
            linter.lint(descriptor, out)



from eip712.model.schema import EIP712SchemaField, EIP712Type
from pydantic import Field

from erc7730.model.base import Model

# ruff: noqa: N815 - camel case field names are tolerated to match schema


class EIP712Schema(Model):
    """
    EIP-712 message schema.
    """

    primaryType: EIP712Type = Field(title="Primary Type", description="The identifier of the schema primary type.")

    types: dict[EIP712Type, list[EIP712SchemaField]] = Field(
        title="Types", description="The schema types reachable from primary type."
    )



from erc7730.common.output import OutputAdder
from erc7730.model.metadata import EnumDefinition
from erc7730.model.paths import DescriptorPath, Field
from erc7730.model.paths.path_ops import descriptor_path_strip_prefix
from erc7730.model.types import Id

ENUMS_PATH = DescriptorPath(elements=[Field(identifier="metadata"), Field(identifier="enums")])


def get_enum(ref: DescriptorPath, enums: dict[Id, EnumDefinition], out: OutputAdder) -> dict[str, str] | None:
    if (enum_id := get_enum_id(ref, out)) is None:
        return None

    if (enum := enums.get(enum_id)) is None:
        return out.error(
            title="Invalid enum reference",
            message=f"""Enum "{enum_id}" does not exist, valid ones are: """ f"{', '.join(enums.keys())}.",
        )
    return enum


def get_enum_id(path: DescriptorPath, out: OutputAdder) -> str | None:
    try:
        tail = descriptor_path_strip_prefix(path, ENUMS_PATH)
    except ValueError:
        return out.error(
            title="Invalid enum reference path",
            message=f"Enums must be defined at {ENUMS_PATH}, {path} is not a valid enum reference.",
        )
    if len(tail.elements) != 1:
        return out.error(
            title="Invalid enum reference path",
            message=f"Enums must be defined directly under {ENUMS_PATH}, deep nesting is not allowed, {path} is not a "
            f"valid enum reference.",
        )
    if not isinstance(element := tail.elements[0], Field):
        return out.error(
            title="Invalid enum reference path",
            message=f"Enums must be defined at {ENUMS_PATH}, array operators are not allowed, {path} is not a valid "
            f"enum reference.",
        )

    return element.identifier



from pathlib import Path

import pytest

from erc7730.common.output import ExceptionsToOutput, SetOutputAdder
from erc7730.model.input.descriptor import InputERC7730Descriptor
from tests.cases import TestCase, case_id

DATA = Path(__file__).resolve().parent / "data"


@pytest.mark.parametrize(
    "testcase",
    [
        TestCase(
            id="format_token_amount_native_currency_address_using_data_path",
            label="using a data path where an address value is expected",
            description="using a literal value where a data path is expected",
            error="expected a 20 bytes, hexadecimal Ethereum address",
        ),
    ],
    ids=case_id,
)
def test_errors_by_reference(testcase: TestCase) -> None:
    """
    Test loading and validating and erroneous ERC-7730 descriptor files, and compare against expected errors.
    """
    input_descriptor_path = DATA / "errors" / f"{testcase.id}.json"
    if (expected_error := testcase.error) is None:
        pytest.fail("Testcase must have an expected error")
    out = SetOutputAdder()
    with ExceptionsToOutput(out):
        InputERC7730Descriptor.load(input_descriptor_path)
    messages = {output.message for output in out.outputs}
    assert any(expected_error in message for message in messages), "Expected error not found in:\n" + "\n".join(
        messages
    )



from eip712.model.schema import EIP712SchemaField

from erc7730.model.abi import Component, Function, InputOutput
from erc7730.model.context import EIP712Schema
from erc7730.model.paths.path_parser import to_path
from erc7730.model.paths.path_schemas import compute_abi_schema_paths, compute_eip712_schema_paths


def test_compute_abi_paths_no_params() -> None:
    abi = Function(name="transfer", inputs=[])
    expected: set[str] = set()
    assert compute_abi_schema_paths(abi) == expected


def test_compute_abi_paths_with_params() -> None:
    abi = Function(
        name="transfer", inputs=[InputOutput(name="to", type="address"), InputOutput(name="amount", type="uint256")]
    )
    expected = {to_path("#.to"), to_path("#.amount")}
    assert compute_abi_schema_paths(abi) == expected


def test_compute_abi_paths_with_slicable_params() -> None:
    abi = Function(
        name="transfer", inputs=[InputOutput(name="to", type="bytes"), InputOutput(name="amount", type="uint256")]
    )
    expected = {to_path("#.to"), to_path("#.to.[]"), to_path("#.amount")}
    assert compute_abi_schema_paths(abi) == expected


def test_compute_abi_paths_with_nested_params() -> None:
    abi = Function(
        name="foo",
        inputs=[
            InputOutput(
                name="bar",
                type="tuple",
                components=[Component(name="baz", type="uint256"), Component(name="qux", type="address")],
            )
        ],
    )
    expected = {to_path("#.bar.baz"), to_path("#.bar.qux")}
    assert compute_abi_schema_paths(abi) == expected


def test_compute_abi_paths_with_multiple_nested_params() -> None:
    abi = Function(
        name="foo",
        inputs=[
            InputOutput(
                name="bar",
                type="tuple",
                components=[
                    Component(name="baz", type="bytes"),
                    Component(name="qux", type="address"),
                    Component(name="nested", type="tuple[]", components=[Component(name="deep", type="string")]),
                ],
            )
        ],
    )
    expected = {
        to_path("#.bar.baz"),
        to_path("#.bar.baz.[]"),
        to_path("#.bar.qux"),
        to_path("#.bar.nested.[]"),
        to_path("#.bar.nested.[].deep"),
    }
    assert compute_abi_schema_paths(abi) == expected


def test_compute_eip712_paths_with_slicable_params() -> None:
    schema = EIP712Schema(
        primaryType="Foo",
        types={"Foo": [EIP712SchemaField(name="bar", type="bytes")]},
    )
    expected = {
        to_path("#.bar"),
        to_path("#.bar.[]"),
    }
    assert compute_eip712_schema_paths(schema) == expected


def test_compute_eip712_paths_with_multiple_nested_params() -> None:
    schema = EIP712Schema(
        primaryType="Foo",
        types={
            "Foo": [
                EIP712SchemaField(name="bar", type="Bar"),
            ],
            "Bar": [
                EIP712SchemaField(name="baz", type="bytes"),
                EIP712SchemaField(name="qux", type="uint256"),
                EIP712SchemaField(name="nested", type="Nested[]"),
            ],
            "Nested": [
                EIP712SchemaField(name="deep", type="uint256"),
            ],
        },
    )
    expected = {
        to_path("#.bar.baz"),
        to_path("#.bar.baz.[]"),
        to_path("#.bar.qux"),
        to_path("#.bar.nested.[]"),
        to_path("#.bar.nested.[].deep"),
    }
    assert compute_eip712_schema_paths(schema) == expected



import os
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from typing import Any, LiteralString, TypeVar

from pydantic import BaseModel, ValidationInfo, WrapValidator
from pydantic_core import PydanticCustomError

from erc7730.common.json import dict_to_json_file, dict_to_json_str, read_json_with_includes

_BaseModel = TypeVar("_BaseModel", bound=BaseModel)


def model_from_json_bytes(data: bytes, model: type[_BaseModel]) -> _BaseModel:
    """Load a Pydantic model from JSON content as an array of bytes."""
    return model.model_validate_json(data, strict=True)


def model_from_json_str(data: str, model: type[_BaseModel]) -> _BaseModel:
    """Load a Pydantic model from JSON content as an array of bytes."""
    return model.model_validate_json(data, strict=True)


def model_from_json_file_with_includes(path: Path, model: type[_BaseModel]) -> _BaseModel:
    """Load a Pydantic model from a JSON file, including references."""
    return model.model_validate(read_json_with_includes(path), strict=False)


def model_from_json_file_with_includes_or_none(path: Path, model: type[_BaseModel]) -> _BaseModel | None:
    """Load a Pydantic model from a JSON file, or None if file does not exist."""
    return model_from_json_file_with_includes(path, model) if os.path.isfile(path) else None


def model_to_json_dict(obj: _BaseModel) -> dict[str, Any]:
    """Serialize a pydantic model into a JSON dict."""
    return obj.model_dump(mode="json", by_alias=True, exclude_none=True)


def model_to_json_str(obj: _BaseModel) -> str:
    """Serialize a pydantic model into a JSON string."""
    return dict_to_json_str(model_to_json_dict(obj))


def model_to_json_file(path: Path, model: _BaseModel) -> None:
    """Write a model to a JSON file, creating parent directories as needed."""
    dict_to_json_file(path, model_to_json_dict(model))


@dataclass(frozen=True)
class ErrorTypeLabel(WrapValidator):
    """
    Wrapper validator that replaces all errors with a simple message "expected a <type label>".

    It is useful for annotating union types where pydantic returns multiple errors for each type it tries, or custom
    base types such as pattern validated strings to get more user-friendly errors.
    """

    def __init__(self, type_label: LiteralString) -> None:
        super().__init__(self._validator(type_label))

    @staticmethod
    def _validator(type_label: LiteralString) -> Callable[[Any, Any, ValidationInfo], Any]:
        def validate(v: Any, next_: Any, ctx: ValidationInfo) -> Any:
            try:
                return next_(v, ctx)
            except Exception:
                raise PydanticCustomError("custom_error", "expected a " + type_label) from None

        return validate
